# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Help functions for Synapse pipelines notebooks."""
import logging
import os
import re
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

import httpx
import jwt

from .._version import VERSION
from ..auth.azure_auth import AzureCredEnvNames, az_connect
from ..auth.cloud_mappings import AzureCloudConfig
from ..common.pkg_config import get_config, get_http_timeout, refresh_config, set_config
from ..common.provider_settings import get_provider_settings
from ..common.utility import mp_ua_header

logger = logging.getLogger(__name__)


class SparkUtilsPlaceHolder:
    """Placeholder for non-spark environments."""

    def __getattr__(self, attrib):
        """Raise not implemented error."""
        raise NotImplementedError(
            "This function can only be used in an Azure Synapse Spark environment."
        )


try:
    from notebookutils import mssparkutils

    _IN_SPARK = True
except ImportError:
    mssparkutils = SparkUtilsPlaceHolder()
    _IN_SPARK = False


__version__ = VERSION
__author__ = "Ian Hellen"

_LINKED_SERVICES_URL = (
    "https://{ws_name}.{synapse_endpoint}/linkedservices?api-version=2020-12-01"
)

_AZ_NAME_PATTERN = re.compile(r"^https://(?P<name>[^.]+)\.(?P<suffix>.*$)")


def is_in_synapse():
    """Return True if running in Synapse Pipeline."""
    return os.environ.get("MMLSPARK_PLATFORM_INFO") == "synapse"


IdentityType = Literal["managed", "service_principal"]


def init_synapse(
    identity_type: IdentityType = "service_principal",
    storage_svc_name: Optional[str] = None,
    tenant_id: Optional[str] = None,
    cloud: str = "global",
):
    """
    Initialize Synapse Spark notebook pipeline.

    Parameters
    ----------
    identity_type : IdentityType, optional
        The authentication type to be used by MSTICPy, by default "managed".
        Options are "managed", "service_principal"
    storage_svc_name : Optional[str], optional
        Override the default storage linked service name, by default None
    tenant_id : Optional[str], optional
        Override the default tenant_id for the Azure authentication, by default None
    cloud : str
        Azure cloud - default is "global".

    Raises
    ------
    RuntimeError
        No suitable linked storage service found.
        Could not mount configuration data container.
        Could not authenticate to Azure.
        Could not retrieve secret from Key Vault.

    """
    if not is_in_synapse():
        print("Not running in Azure Synapse environment.")
        logger.info("Not running in Azure Synapse environment.")
        return
    mp_spark = MPSparkUtils()
    storage_svc = mp_spark.get_storage_service(storage_svc_name)
    if not storage_svc:
        logger.error("No suitable linked storage service found.")
        raise RuntimeError("No suitable linked storage service found.")
    print(f"Using linked storage service {storage_svc.name}")
    logger.info("Using linked storage service %s", storage_svc.name)
    print("Mounting storage...", end=" ")
    stor_acct_match = re.match(_AZ_NAME_PATTERN, storage_svc.url)
    mount_success = mount_container(
        store_acct_name=stor_acct_match["name"] if stor_acct_match else "",
        container=mp_spark.container,
        mount_path=mp_spark.mount_point,
        linked_service=storage_svc.name,
    )
    if not mount_success:
        raise RuntimeError("Could not mount configuration data container.")
    print(f"Mounted {storage_svc.url}/{mp_spark.container} at {mp_spark.config_path}")
    logger.info(
        "Mounted %s/%s at %s", storage_svc.url, mp_spark.container, mp_spark.config_path
    )

    # Set msticpy settings
    print("Configuring msticpy...", end=" ")
    _configure_mp_settings(mp_spark)
    logger.info("msticpyconfig settings updated")
    print("settings,", end=" ")

    # Set up authentication parameters
    if identity_type == "managed":
        _set_mp_azure_settings(auth_method="msi", cloud=cloud)
        _set_msi_client_id(mp_spark=mp_spark, tenant_id=tenant_id)
        print("using managed identity,", end=" ")
        logger.info("Using managed identity")
    else:
        # Publish SP creds from KeyVault
        _set_mp_azure_settings(auth_method="env", cloud=cloud)
        _set_azure_env_creds(mp_spark=mp_spark, tenant_id=tenant_id)
        print("using service principal,", end=" ")
        logger.info("Using service principal")
    print()

    # authenticate with MSI or SP to linked KV
    creds = az_connect()
    if not creds:
        logger.error("Could not authenticate to Azure.")
        raise RuntimeError("Could not authenticate to Azure.")
    print("Testing authentication...", end=" ")

    # check key retrieval
    if not _check_kv_key_retrieval():
        logger.error("Could not retrieve secret from Key Vault.")
        raise RuntimeError("Could not retrieve secret from Key Vault.")
    print("Key Vault success.")
    print("Synapse initialization successful")
    logger.info("Synapse initialization successful")


def current_mounts() -> Dict[str, str]:
    """Return dictionary of current Synapse mount points."""
    return {mnt.mountPoint: mnt.source for mnt in mssparkutils.fs.mounts()}


def mount_container(
    store_acct_name: str,
    container: str,
    mount_path: str,
    linked_service: str,
    folder: str = "",
    cloud: str = "global",
) -> bool:
    """
    Mount Azure file container to Synapse file system.

    Parameters
    ----------
    store_acct_name : str
        Storage account
    container : str
        Container name to mount
    mount_path : str
        Local path to mount
    linked_service : str
        Storage linked service name
    folder : str, optional
        Subfolder of container, if any, by default ""
    cloud : str, optional
        Azure Cloud, by default "global"

    Returns
    -------
    bool
        True if mounting successful or if mount point already connected.
        False if mounting failed or mount point connected to a different
        storage location.

    Raises
    ------
    NotImplementedError
        If using a cloud other than Azure Global (public cloud).

    """
    if cloud != "global":
        raise NotImplementedError("Only supports Azure 'global' cloud currently.")

    mount_path = mount_path if mount_path.startswith("/") else f"/{mount_path}"
    existing_mounts = current_mounts()

    storage_root = "dfs.core.windows.net/"
    folder = f"{folder}/" if folder else ""

    storage_url = f"abfss://{container}@{store_acct_name}.{storage_root}{folder}"
    logger.info("Mounting %s as %s", storage_url, mount_path)
    if mount_path in existing_mounts:
        print(f"path '{mount_path}' already mounted to {existing_mounts[mount_path]}")
        return existing_mounts[mount_path] == storage_url

    return mssparkutils.fs.mount(
        storage_url, mount_path, {"linkedService": linked_service}
    )


# pylint: disable=too-few-public-methods
class LinkedService:
    """Azure Synapse Linked Service settings."""

    _TYPE_PROPS = "typeProperties"

    class ServiceTypes:
        """Enumeration of linked service type names."""

        AzureBlobFS = "AzureBlobFS"
        AzureSqlDW = "AzureSqlDW"
        AzureFileStorage = "AzureFileStorage"
        AzureBlobStorage = "AzureBlobStorage"
        AzureKeyVault = "AzureKeyVault"

    def __init__(self, **kwargs):
        """Initialize Linked Service instance."""
        self.res_id: str = kwargs.get("id")
        self.name: str = kwargs.get("name")
        self.entry_type: str = kwargs.get("type")
        self.etag: str = kwargs.get("etag")
        self.properties: Dict[str, Union[str, Dict[str, Union[Dict, str]]]] = (
            kwargs.get("properties", {})
        )

    @property
    def svc_type(self) -> str:
        """Return type of linked service."""
        return self.properties.get("type")  # type: ignore

    def __getattr__(self, name: str):
        """Return sub property `name` from properties or typeProperties."""
        if name in self.properties:
            return self.properties[name]
        if name in self.properties.get(self._TYPE_PROPS, {}):
            return self.properties.get(self._TYPE_PROPS, {})[name]  # type: ignore
        raise AttributeError(f"Unknown attribute {name}")

    def __repr__(self):
        """Return string representation of instance."""
        return f"LinkedService(name='{self.name}', type='{self.svc_type}')"

    @property
    def azure_name(self):
        """Return Azure resource name."""
        url = self.properties.get(self._TYPE_PROPS, {}).get("url")
        if not url:
            url = self.properties.get(self._TYPE_PROPS, {}).get("baseUrl")
        if not url:
            return None
        az_name = re.search(_AZ_NAME_PATTERN, url)
        return az_name["name"]


# Note storage linked service is generated by Synapse workspace creation process
# pylint: disable=too-few-public-methods
class SynapseName:
    """Name mapping to default values."""

    storage_account_prefix = (
        "adlsforsentinel"  # + last 7 digit of Sentinel workspace id;
    )
    key_vault_name_prefix = "kvforsentinel"  # + last 7 digit of ws I’d;
    kv_linked_service = "Akvlink"
    sp_client_id_name = "clientid"
    # pylint: disable=line-too-long
    # [SuppressMessage("Microsoft.Security", "CS002:SecretInNextLine", Justification="This is an enum of property names")]
    sp_client_sec_name = "clientsecret"
    container = "sentinelfiles"

    def __init__(self, workspace_id: str):
        """Initialize the Synapse Name class."""
        self._ws_id = workspace_id

    @property
    def storage_account(self) -> str:
        """Return default storage account name."""
        return f"{self.storage_account_prefix}{self._ws_id[-7:]}"

    @property
    def key_vault(self) -> str:
        """Return default Key Vault name."""
        return f"{self.key_vault_name_prefix}{self._ws_id[-7:]}"


class MPSparkUtils:
    """MSTICPy Spark Synapse Utility class."""

    _DEF_MOUNT_POINT = "msticpy"

    def __init__(
        self, mount_point: Optional[str] = None, container: Optional[str] = None
    ):
        """Initialize MPSparkUtils class."""
        self.linked_services = [
            LinkedService(**props)
            for props in _fetch_linked_services(self.workspace_name)
        ]
        self._get_workspace_ids()
        self.mount_point = mount_point or self._DEF_MOUNT_POINT
        self.container = container or SynapseName.container

    @property
    def workspace_name(self) -> str:
        """Return the Synapse workspace name."""
        return mssparkutils.env.getWorkspaceName()

    @property
    def fs_mounts(self) -> Dict[str, str]:
        """Return a dictionary of mount points and targets."""
        return {mnt.mountPoint: mnt.source for mnt in mssparkutils.fs.mounts()}

    @property
    def job_id(self):
        """Return current job ID."""
        return mssparkutils.env.getJobId()

    @property
    def config_path(self):
        """Return mount path for MSTICPy config."""
        return Path(f"/synfs/{self.job_id}/{self.mount_point}")

    def get_service_of_type(self, svc_type: str) -> Optional[LinkedService]:
        """
        Return the first linked service of specific `svc_type`.

        Parameters
        ----------
        svc_type : str
            Service Type (AzureBlobFS, AzureBlobStorage, AzureSqlDW,
            AzureKeyVault)

        Returns
        -------
        Optional[LinkedService]
            LinkedService instance for the `svc_type`.

        """
        return next(
            iter(
                lnk_svc
                for lnk_svc in self.linked_services
                if lnk_svc.svc_type == svc_type
            ),
            None,
        )

    def get_all_services_of_type(self, svc_type) -> List[LinkedService]:
        """
        Return list of Linked services of `svc_type`.

        Parameters
        ----------
        svc_type : str
            Service Type (AzureBlobFS, AzureBlobStorage, AzureSqlDW,
            AzureKeyVault)

        Returns
        -------
        List[LinkedService]
            List of LinkedService instances of type `svc_type`.

        """
        return [
            lnk_svc for lnk_svc in self.linked_services if lnk_svc.svc_type == svc_type
        ]

    def get_ws_default_storage(self) -> Optional[LinkedService]:
        """
        Return default storage linked service.

        Returns
        -------
        Optional[LinkedService]
            Default storage service.

        """
        try:
            return next(
                iter(
                    lnk_svc
                    for lnk_svc in self.linked_services
                    if lnk_svc.svc_type == LinkedService.ServiceTypes.AzureBlobFS
                    and "WorkspaceDefaultStorage" in lnk_svc.name
                )
            )
        except StopIteration:
            return None

    def get_service(self, svc_name: str) -> Optional[LinkedService]:
        """
        Return named linked service.

        Returns
        -------
        Optional[LinkedService]
            Named service service.

        """
        try:
            return next(
                iter(
                    lnk_svc
                    for lnk_svc in self.linked_services
                    if lnk_svc.name == svc_name
                )
            )
        except StopIteration:
            return None

    def get_storage_service(
        self, linked_svc_name: Optional[str] = None
    ) -> LinkedService:
        """Return linked storage service (named) or default storage."""
        storage_svc: Optional[LinkedService] = None
        if linked_svc_name:
            storage_svc = self.get_service(svc_name=linked_svc_name)
        if not storage_svc:
            storage_svc = self.get_ws_default_storage()
            if storage_svc is None:
                storage_svc = self.get_service_of_type(
                    svc_type=LinkedService.ServiceTypes.AzureBlobFS
                )
        if not storage_svc:
            raise TypeError("No linked storage service found.")
        return storage_svc

    def get_kv_secret(self, secret_name: str) -> str:
        """Return secret from linked service."""
        kv_service = self.get_service_of_type(LinkedService.ServiceTypes.AzureKeyVault)
        if not kv_service:
            raise ValueError("No linked Key Vault service found.")
        return mssparkutils.credentials.getSecret(
            kv_service.azure_name, secret_name, kv_service.name
        )

    def _get_workspace_ids(self):
        """Retrieve Synapse Managed Identity details from token."""
        ws_token = mssparkutils.credentials.getToken("Synapse")
        alg = jwt.get_unverified_header(ws_token)["alg"]
        decoded_token = jwt.decode(
            ws_token, algorithms=[alg], options={"verify_signature": False}
        )
        self.application_id = decoded_token["appid"]
        self.tenant_id = decoded_token["tid"]
        self.object_id = decoded_token["oid"]
        self.resource_id = decoded_token.get("xms_mirid")
        logger.info(
            "Workspace IDs retrieved: app_id: %s, tenant_id: %s, obj_id: %s, res_id: %s",
            self.application_id,
            self.tenant_id,
            self.object_id,
            self.resource_id,
        )


def _fetch_linked_services(ws_name: str):
    """Fetch list of linked services via Azure Synapse API."""
    az_cloud_config = AzureCloudConfig()
    synapse_endpoint = az_cloud_config.suffixes.get("synapseAnalytics")
    token = mssparkutils.credentials.getToken("Synapse")
    req_headers = {"Authorization": f"Bearer {token}", **mp_ua_header()}

    resp = httpx.get(
        _LINKED_SERVICES_URL.format(ws_name=ws_name, synapse_endpoint=synapse_endpoint),
        headers=req_headers,
        timeout=get_http_timeout(),
    )
    return resp.json().get("value")


def _set_azure_env_creds(mp_spark: MPSparkUtils, tenant_id: Optional[str] = None):
    """Publish Service Principal credentials to environment variables."""
    os.environ[AzureCredEnvNames.AZURE_TENANT_ID] = tenant_id or mp_spark.tenant_id
    client_id = mp_spark.get_kv_secret(SynapseName.sp_client_id_name)
    if not client_id:
        raise LookupError("Could not obtain client_id from Key Vault.")
    os.environ[AzureCredEnvNames.AZURE_CLIENT_ID] = client_id
    client_secret = mp_spark.get_kv_secret(SynapseName.sp_client_sec_name)
    if not client_secret:
        raise LookupError("Could not obtain client_secret from Key Vault.")
    os.environ[AzureCredEnvNames.AZURE_CLIENT_SECRET] = client_secret
    logger.info(
        "Setting authentication env var %s=%s",
        AzureCredEnvNames.AZURE_TENANT_ID,
        tenant_id or mp_spark.tenant_id,
    )
    logger.info(
        "Setting authentication env var %s=%s",
        AzureCredEnvNames.AZURE_CLIENT_ID,
        client_id,
    )
    logger.info(
        "Setting authentication env var %s=%s",
        AzureCredEnvNames.AZURE_CLIENT_SECRET,
        "*" * len(client_secret),
    )


def _set_msi_client_id(mp_spark: MPSparkUtils, tenant_id: Optional[str] = None):
    """Publish Service Principal credentials to environment variables."""
    os.environ[AzureCredEnvNames.AZURE_TENANT_ID] = tenant_id or mp_spark.tenant_id
    os.environ[AzureCredEnvNames.AZURE_CLIENT_ID] = mp_spark.application_id
    logger.info(
        "Setting authentication env var %s=%s",
        AzureCredEnvNames.AZURE_TENANT_ID,
        tenant_id or mp_spark.tenant_id,
    )
    logger.info(
        "Setting authentication env var %s=%s",
        AzureCredEnvNames.AZURE_CLIENT_ID,
        mp_spark.application_id,
    )


def _set_mp_azure_settings(auth_method: str, cloud: str = "global"):
    """Configure in-memory settings for MP Azure authn_methods."""
    az_settings = get_config("Azure", {})
    if not az_settings:
        set_config(
            "Azure",
            {
                "auth_methods": [auth_method],
                "cloud": cloud,
            },
        )
        logger.info("Added Azure configuration to config.")
    else:
        curr_methods = az_settings.get("auth_methods", [])
        if auth_method in curr_methods:
            curr_methods.remove(auth_method)
        az_settings["auth_methods"] = [auth_method, *curr_methods]
        logger.info("Added Azure configuration authentication methods.")


def _check_kv_key_retrieval() -> bool:
    """Check that we are able to get a key from Key Vault."""
    ti_settings = get_provider_settings("TIProviders")
    prov_settings = next(
        iter(
            settings
            for settings in ti_settings.values()
            if settings.args and "AuthKey" in settings.args
        ),
        None,
    )
    if prov_settings is None:
        logger.warning("No provider settings to test Key Vault connection.")
        return False
    value = prov_settings.args.get("AuthKey")
    return value is not None


def _configure_mp_settings(mp_spark: MPSparkUtils):
    """Set msticpyconfig and GeoIPLite dbfolder paths."""
    mp_path = mp_spark.config_path.joinpath("msticpyconfig.yaml")
    if not mp_path.is_file():
        raise RuntimeError("No msticpyconfig.yaml found at", str(mp_spark.config_path))
    os.environ["MSTICPYCONFIG"] = str(mp_path)
    os.environ["MSTICPYHOME"] = str(mp_spark.config_path)
    before_config_keys = len(get_config())
    refresh_config()
    if len(get_config()) > before_config_keys:
        sections = ", ".join(get_config().keys())
        logger.warning(
            "Possible problem loading msticpyconfig. Sections loaded: %s",
            sections,
        )
    geolite_settings = get_config("OtherProviders.GeoIPLite.Args", None)
    if geolite_settings:
        set_config("OtherProviders.GeoIPLite.Args.DBFolder", str(mp_spark.config_path))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Context Providers Subpackage."""
from __future__ import annotations

from typing import Any

from ..common.utility import ImportPlaceholder
from ..lazy_importer import lazy_import
from .vtlookupv3 import VT3_AVAILABLE

vtlookupv3: Any
if VT3_AVAILABLE:
    from .vtlookupv3 import vtlookupv3
else:
    # vtlookup3 will not load if vt package not installed
    vtlookupv3 = ImportPlaceholder(
        "vtlookupv3",
        ["vt-py", "vt-graph-api", "nest_asyncio"],
    )


_LAZY_IMPORTS: set[str] = {
    "msticpy.context.geoip.GeoLiteLookup",
    "msticpy.context.geoip.IPStackLookup",
    "msticpy.context.tilookup.TILookup",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
import random
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import ipywidgets as widgets
from deprecated.sphinx import deprecated
from IPython.display import HTML, display
from ipywidgets import Layout

from .._version import VERSION
from .core import IPyDisplayMixin

__version__ = VERSION
__author__ = "Ian Hellen"


class SelectItem(IPyDisplayMixin):
    """
    Selection list from list or dict.

    Attributes
    ----------
    item_action : Callable[..., Optional[Tuple]]
        Action to call for each selection.

    """

    # pylint: disable=too-many-arguments, too-few-public-methods
    def __init__(
        self,
        description: str = "Select an item",
        options: Union[List[str], Dict[str, Any]] = None,
        action: Callable[..., Optional[Tuple]] = None,
        value: str = "",
        **kwargs,
    ):
        """
        Select an item from a list or dict.

        Parameters
        ----------
        description : str, optional
            The widget label to display.
            (the default is 'Select an item')
        options : Union[List[str], Dict[str, Any]]
            Either:
            A `list` of items to select from (the default is None)
            A `dict` of items to select from. When using `item_dict`
            the values are displayed as the selectable items and key
            corresponding to the selected value is set as the `value`
            property.
        action : Callable[..., Optional[Tuple[...]]], optional
            function to call when item selected (passed a single
            parameter - the value of the currently selected item)
            (the default is None).
            If the function returns one or a tuple of displayable objects
            these will be displayed.
        value : str, optional
            A default value to pre-populate the filter with.

        Other Parameters
        ----------------
        item_list : List[str], optional
            A `list` of items to select from (the default is None)
        item_dict : Dict[str, str], optional
            A `dict` of items to select from. When using `item_dict`
            the keys are displayed as the selectable items and value
            corresponding to the selected key is set as the `value`
            property.
            (the default is None)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)
        height : str, optional
            Selection list height (the default is '100px')
        width : str, optional
            Selection list width (the default is '50%')
        display_filter : bool, optional
            Whether to display item filter (the default is True)

        """
        self.options = options or kwargs.pop("item_list", [])
        self.options = self.options or kwargs.pop("item_dict", {})

        if not self.options:
            raise ValueError("No options supplied for SelectItem.")
        auto_display = kwargs.pop("auto_display", False)
        height: str = kwargs.pop("height", "100px")
        width: str = kwargs.pop("width", "50%")
        display_filter: bool = kwargs.pop("display_filter", True)

        self.def_value = value
        if self.def_value not in self.options:
            if isinstance(self.options, list):
                self.def_value = next(iter(self.options))
            else:
                self.def_value = next(iter(self.options.values()))

        self._wgt_select = widgets.Select(
            value=self.def_value,
            options=self._get_filtered_options(),
            description=description,
            layout=Layout(width=width, height=height),
            style={"description_width": "initial"},
        )
        self._w_display_details = widgets.Checkbox(
            value=True,
            description="Display details",
        )
        self._display_filter = display_filter
        if display_filter:
            self._w_filter = widgets.Text(
                value=self.label,
                description="Filter:",
                style={"description_width": "initial"},
            )

            # set up observer callbacks
            self._w_filter.observe(self._filter_options, names="value")
        self._wgt_select.observe(self._select_item, names="value")

        if action:
            self._w_display_details.observe(self._run_action, names="value")
        self.item_action = action

        # setup to use updatable display objects
        rand_id = random.randint(0, 999999)  # nosec
        self._output_id = f"{self.__class__.__name__}_{rand_id}"
        self._disp_elems: List[Any] = []

        if auto_display:
            self.display()

    @property
    def layout(self):
        """Return underlying widget collection."""
        wgt_list = []
        if self._display_filter:
            wgt_list.append(self._w_filter)
        wgt_list.append(self._wgt_select)
        if self.item_action:
            wgt_list.append(self._w_display_details)
        return widgets.VBox(wgt_list)

    @property
    def value(self):
        """Return the currently selected item."""
        return self._wgt_select.value

    @value.setter
    def value(self, value):
        """Set current selected value of widget."""
        if value in self.options:
            self._wgt_select.label = value
        if isinstance(self.options, dict) and value in self.options.values():
            self._wgt_select.value = value

    @property
    def label(self):
        """Return current display item."""
        return self._wgt_select.label

    def display(self):
        """Display the interactive widget."""
        super().display()
        display(HTML("<hr>"))
        self._show_top_item()

    def _select_item(self, selection):
        """Run action if available."""
        del selection
        if self.item_action is not None:
            self._run_action()

    def _filter_options(self, change):
        """Filter the alert list by substring."""
        if change is None or "new" not in change:
            return
        self._wgt_select.options = self._get_filtered_options(change["new"])

    def _get_filtered_options(
        self, substring: str = ""
    ) -> List[Union[str, Tuple[str, str]]]:
        """Return optionally filtered list of option tuples."""
        if self.options is None:
            return []
        if isinstance(self.options, list):
            return [val for val in self.options if substring.casefold() in val.lower()]
        return [
            (lab, val)
            for lab, val in self.options.items()
            if substring.casefold() in lab.lower()
        ]

    def _run_action(self, change=None):
        """Run any action function and display details, if any."""
        del change
        output_objs = None
        if self._w_display_details.value:
            output_objs = self.item_action(self.value)
        if output_objs is None:
            self._clear_display()
            return
        if not isinstance(output_objs, (tuple, list)):
            output_objs = [output_objs]
        display_objs = dict(enumerate(self._disp_elems))
        for idx, out_obj in enumerate(output_objs):
            if idx not in display_objs:
                self._disp_elems.append(
                    display(out_obj, display_id=f"{self._output_id}_{idx}")
                )
            else:
                self._disp_elems[idx].update(out_obj)

    def _clear_display(self):
        """Clear any current details."""
        if not self._disp_elems:
            return
        for disp_obj in self._disp_elems:
            disp_obj.update(HTML(""))

    def _show_top_item(self):
        """Run action on the first item by default."""
        if self.item_action is not None and self.value is not None:
            self._run_action()


@deprecated(
    reason="Superceded by SelectItem. Will be removed in v2.0.0.", version="0.5.2"
)
class SelectString(SelectItem):
    """Selection list from list or dict."""

    # pylint: disable=too-many-arguments, too-few-public-methods
    def __init__(
        self,
        description: str = "Select an item",
        item_list: List[str] = None,
        action: Callable[..., None] = None,
        item_dict: Dict[str, str] = None,
        auto_display: bool = False,
        height: str = "100px",
        width: str = "50%",
        display_filter: bool = True,
    ):
        """
        Select an item from a list or dict.

        Parameters
        ----------
        description : str, optional
            The widget label to display.
            (the default is 'Select an item')
        item_list : List[str], optional
            A `list` of items to select from (the default is None)
        item_dict : Dict[str, str], optional
            A `dict` of items to select from. When using `item_dict`
            the keys are displayed as the selectable items and value
            corresponding to the selected key is set as the `value`
            property.
            (the default is None)
        action : Callable[..., None], optional
            function to call when item selected (passed a single
            parameter - the value of the currently selected item)
            (the default is None)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)
        height : str, optional
            Selection list height (the default is '100px')
        width : str, optional
            Selection list width (the default is '50%')
        display_filter : bool, optional
            Whether to display item filter (the default is True)

        """
        self._w_output = widgets.Output(layout={"border": "1px solid black"})

        super().__init__(
            description=description,
            item_list=item_list,
            item_dict=item_dict,
            action=action,
            auto_display=auto_display,
            height=height,
            width=width,
            display_filter=display_filter,
        )

    def _run_action(self, change=None):
        del change
        self._w_output.clear_output()
        with self._w_output:
            self.item_action(self.value)

    def display(self):
        """Display the interactive widget."""
        self._show_top_item()
        wgt_list = []
        if self._display_filter:
            wgt_list.append(self._w_filter)
        wgt_list.append(self._wgt_select)
        if self.item_action:
            wgt_list.append(self._w_output)
        display(widgets.VBox(wgt_list))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query Provider additional connection methods."""
from __future__ import annotations

import asyncio
import logging
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from itertools import tee
from typing import TYPE_CHECKING, Any, Protocol

import nest_asyncio
import pandas as pd
from tqdm.auto import tqdm
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyDataQueryError
from ...common.utility.ipython import is_ipython
from ..drivers.driver_base import DriverBase, DriverProps

if TYPE_CHECKING:
    from datetime import datetime

    from .query_source import QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"

logger: logging.Logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods, unnecessary-ellipsis
class QueryProviderProtocol(Protocol):
    """Protocol for required properties of QueryProvider class."""

    driver_class: type[DriverBase]
    _driver_kwargs: dict[str, Any]
    _additional_connections: dict[str, DriverBase]
    _query_provider: DriverBase

    @staticmethod
    @abstractmethod
    def _get_query_options(
        params: dict[str, Any],
        kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """Return any kwargs not already in params."""


# pylint: disable=super-init-not-called
class QueryProviderConnectionsMixin(QueryProviderProtocol):
    """Mixin additional connection handling QueryProvider class."""

    @staticmethod
    @abstractmethod
    def _get_query_options(
        params: dict[str, Any],
        kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """Return any kwargs not already in params."""

    def exec_query(self: Self, query: str, **kwargs) -> pd.DataFrame | str | None:
        """
        Execute simple query string.

        Parameters
        ----------
        query : str
            [description]
        use_connections : Union[str, list[str]]

        Other Parameters
        ----------------
        query_options : dict[str, Any]
            Additional options passed to query driver.
        kwargs : dict[str, Any]
            Additional options passed to query driver.

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results - a DataFrame if successful
            or a KqlResult if unsuccessful.

        """
        query_options: dict[str, Any] = kwargs.pop("query_options", {}) or kwargs
        query_source: QuerySource | None = kwargs.pop("query_source", None)

        logger.info("Executing query '%s...'", query[:40])
        logger.debug("Full query: %s", query)
        logger.debug("Query options: %s", query_options)
        if not self._additional_connections:
            return self._query_provider.query(
                query,
                query_source=query_source,
                **query_options,
            )
        return self._exec_additional_connections(query, **kwargs)

    def add_connection(
        self: Self,
        connection_str: str | None = None,
        alias: str | None = None,
        **kwargs,
    ) -> None:
        """
        Add an additional connection for the query provider.

        Parameters
        ----------
        connection_str : Optional[str], optional
            Connection string for the provider, by default None
        alias : Optional[str], optional
            Alias to use for the connection, by default None

        Other Parameters
        ----------------
        kwargs : dict[str, Any]
            Other connection parameters passed to the driver.

        Notes
        -----
        Some drivers may accept types other than strings for the
        `connection_str` parameter.

        """
        # create a new instance of the driver class
        new_driver: DriverBase = self.driver_class(**(self._driver_kwargs))
        # connect
        new_driver.connect(connection_str=connection_str, **kwargs)
        # add to collection
        driver_key: str = alias or str(len(self._additional_connections))
        self._additional_connections[driver_key] = new_driver

    def list_connections(self: Self) -> list[str]:
        """
        Return a list of current connections.

        Returns
        -------
        list[str]
            The alias and connection string for each connection.

        """
        add_connections: list[str] = [
            f"{alias}: {driver.current_connection}"
            for alias, driver in self._additional_connections.items()
        ]
        return [f"Default: {self._query_provider.current_connection}", *add_connections]

    # pylint: disable=too-many-locals
    def _exec_additional_connections(
        self: Self,
        query: str,
        *,
        progress: bool = True,
        retry_on_error: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Return results of query run query against additional connections.

        Parameters
        ----------
        query : str
            The query to execute.
        progress: bool, optional
            Show progress bar, by default True
        retry_on_error: bool, optional
            Retry failed queries, by default False
        **kwargs : dict[str, Any]
            Additional keyword arguments to pass to the query method.

        Returns
        -------
        pd.DataFrame
            The concatenated results of the query executed against all connections.

        Notes
        -----
        This method executes the specified query against all additional connections
        added to the query provider.
        If the driver supports threading or async execution, the per-connection
        queries are executed asynchronously.
        Otherwise, the queries are executed sequentially.

        """
        # Add the initial connection
        query_tasks: dict[str, partial[pd.DataFrame | str | None]] = {
            self._query_provider.current_connection
            or "0": partial(
                self._query_provider.query,
                query,
                **kwargs,
            ),
        }
        # add the additional connections
        query_tasks.update(
            {
                name: partial(connection.query, query, **kwargs)
                for name, connection in self._additional_connections.items()
            },
        )

        logger.info("Running queries for %s connections.", len(query_tasks))
        # Run the queries threaded if supported
        if self._query_provider.get_driver_property(DriverProps.SUPPORTS_THREADING):
            logger.info("Running threaded queries.")
            event_loop: asyncio.AbstractEventLoop = _get_event_loop()
            max_workers: int = self._query_provider.get_driver_property(
                DriverProps.MAX_PARALLEL,
            )
            return event_loop.run_until_complete(
                self._exec_queries_threaded(
                    query_tasks,
                    progress=progress,
                    retry=retry_on_error,
                    max_workers=max_workers,
                ),
            )

        # standard synchronous execution
        logger.info(
            "Running query for %d connections.",
            len(self._additional_connections),
        )
        return self._exec_synchronous_queries(
            progress=progress,
            query_tasks=query_tasks,
        )

    def _exec_split_query(
        self: Self,
        split_by: str,
        query_source: QuerySource,
        query_params: dict[str, Any],
        *,
        progress: bool = True,
        retry_on_error: bool = False,
        debug: bool = False,
        **kwargs,
    ) -> pd.DataFrame | str | None:
        """
        Execute a query that is split into multiple queries.

        Parameters
        ----------
        split_by : str
            The time interval to split the query by.
        query_source : QuerySource
            The query to execute.
        query_params : dict[str, Any]
            The parameters to pass to the query.

        Other Parameters
        ----------------
        debug: bool, optional
            Return queries to be executed rather than execute them, by default False
        progress: bool, optional
            Show progress bar, by default True
        retry_on_error: bool, optional
            Retry failed queries, by default False
        **kwargs : dict[str, Any]
            Additional keyword arguments to pass to the query method.

        Returns
        -------
        pd.DataFrame
            The concatenated results of the query executed against all connections.

        Notes
        -----
        This method executes the time-chunks of the split query.
        If the driver supports threading or async execution, the sub-queries are
        executed asynchronously. Otherwise, the queries are executed sequentially.

        """
        start: datetime | None = query_params.pop("start", None)
        end: datetime | None = query_params.pop("end", None)
        if not (start and end):
            logger.warning("Cannot split a query with no 'start' and 'end' parameters")
            return None

        split_queries: dict[tuple[datetime, datetime], str] = (
            self._create_split_queries(
                query_source=query_source,
                query_params=query_params,
                start=start,
                end=end,
                split_by=split_by,
            )
        )
        if debug:
            return "\n\n".join(
                f"{start}-{end}\n{query}"
                for (start, end), query in split_queries.items()
            )

        query_tasks: dict[str, partial[pd.DataFrame | str | None]] = (
            self._create_split_query_tasks(
                query_source,
                query_params,
                split_queries,
                **kwargs,
            )
        )
        # Run the queries threaded if supported
        if self._query_provider.get_driver_property(DriverProps.SUPPORTS_THREADING):
            logger.info("Running threaded queries.")
            event_loop: asyncio.AbstractEventLoop = _get_event_loop()
            max_workers: int = self._query_provider.get_driver_property(
                DriverProps.MAX_PARALLEL,
            )
            return event_loop.run_until_complete(
                self._exec_queries_threaded(
                    query_tasks,
                    progress=progress,
                    retry=retry_on_error,
                    max_workers=max_workers,
                ),
            )

        # or revert to standard synchronous execution
        return self._exec_synchronous_queries(
            progress=progress,
            query_tasks=query_tasks,
        )

    def _create_split_query_tasks(
        self: Self,
        query_source: QuerySource,
        query_params: dict[str, Any],
        split_queries: dict[tuple[datetime, datetime], str],
        **kwargs,
    ) -> dict[str, partial[pd.DataFrame | str | None]]:
        """Return dictionary of partials to execute queries."""
        # Retrieve any query options passed (other than query params)
        query_options: dict[str, Any] = self._get_query_options(query_params, kwargs)
        logger.info("query_options: %s", query_options)
        logger.info("kwargs: %s", kwargs)
        if "time_span" in query_options:
            del query_options["time_span"]
        return {
            f"{start}-{end}": partial(
                self.exec_query,
                query=query_str,
                query_source=query_source,
                time_span={"start": start, "end": end},
                **query_options,
            )
            for (start, end), query_str in split_queries.items()
        }

    @staticmethod
    def _exec_synchronous_queries(
        *,
        progress: bool,
        query_tasks: dict[str, Any],
    ) -> pd.DataFrame:
        logger.info("Running queries sequentially.")
        results: list[pd.DataFrame] = []
        if progress:
            query_iter = tqdm(query_tasks.items(), unit="sub-queries", desc="Running")
        else:
            query_iter = query_tasks.items()
        for con_name, query_task in query_iter:
            try:
                results.append(query_task())
            except MsticpyDataQueryError:
                logger.info("Query %s failed.", con_name)
        if results:
            return pd.concat(results)

        logger.warning("All queries failed.")
        return pd.DataFrame()

    def _create_split_queries(
        self,
        query_source: QuerySource,
        query_params: dict[str, Any],
        start: datetime,
        end: datetime,
        split_by: str,
    ) -> dict[tuple[datetime, datetime], str]:
        """Return separate queries for split time ranges."""
        try:
            if split_by.strip().endswith("H"):
                split_by = split_by.replace("H", "h")
            split_delta = pd.Timedelta(split_by)
        except ValueError:
            split_delta = pd.Timedelta("1D")
        logger.info("Using split delta %s", split_delta)

        ranges: list[tuple[datetime, datetime]] = _calc_split_ranges(
            start,
            end,
            split_delta,
        )

        split_queries: dict[tuple[datetime, datetime], str] = {
            (q_start, q_end): query_source.create_query(
                formatters=self._query_provider.formatters,
                start=q_start,
                end=q_end,
                **query_params,
            )
            for q_start, q_end in ranges
        }
        logger.info("Split query into %s chunks", len(split_queries))
        return split_queries

    @staticmethod
    async def _exec_queries_threaded(
        query_tasks: dict[str, partial],
        *,
        progress: bool = True,
        retry: bool = False,
        max_workers: int = 4,
    ) -> pd.DataFrame:
        """Return results of multiple queries run as threaded tasks."""
        logger.info("Running threaded queries for %d connections.", len(query_tasks))

        event_loop: asyncio.AbstractEventLoop = _get_event_loop()

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # add the additional connections
            thread_tasks: dict[str, asyncio.Future[pd.DataFrame]] = {
                query_id: event_loop.run_in_executor(executor, query_func)
                for query_id, query_func in query_tasks.items()
            }
            results: list[pd.DataFrame] = []
            failed_tasks_ids: list[str] = []
            if progress:
                task_iter = tqdm(
                    asyncio.as_completed(thread_tasks.values()),
                    unit="sub-queries",
                    desc="Running",
                )
            else:
                task_iter = asyncio.as_completed(thread_tasks.values())
            ids_and_tasks = dict(zip(thread_tasks, task_iter))
            for query_id, thread_task in ids_and_tasks.items():
                try:
                    result: pd.DataFrame | str | None = await thread_task
                    logger.info("Query task '%s' completed successfully.", query_id)
                    results.append(result)
                except Exception:  # pylint: disable=broad-exception-caught
                    logger.warning(
                        "Query task '%s' failed with exception",
                        query_id,
                    )
                    # Reusing thread task would result in:
                    # RuntimeError: cannot reuse already awaited coroutine
                    # A new task should be queued
                    failed_tasks_ids.append(query_id)

        # Sort the results by the order of the tasks
        results = [result for _, result in sorted(zip(thread_tasks, results))]

        if retry and failed_tasks_ids:
            failed_results: pd.DataFrame = (
                await QueryProviderConnectionsMixin._exec_queries_threaded(
                    {
                        failed_tasks_id: query_tasks[failed_tasks_id]
                        for failed_tasks_id in failed_tasks_ids
                    },
                    progress=progress,
                    retry=False,
                    max_workers=max_workers,
                )
            )
            if not failed_results.empty:
                results.append(failed_results)
        if results:
            return pd.concat(results, ignore_index=True)

        logger.warning("All queries failed.")
        return pd.DataFrame()


def _get_event_loop() -> asyncio.AbstractEventLoop:
    """Return the current event loop, or create a new one."""
    try:
        if is_ipython():
            nest_asyncio.apply()
        loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def _calc_split_ranges(
    start: datetime,
    end: datetime,
    split_delta: pd.Timedelta,
) -> list[tuple[datetime, datetime]]:
    """Return a list of time ranges split by `split_delta`."""
    # Use pandas date_range and split the result into 2 iterables
    s_ranges, e_ranges = tee(pd.date_range(start, end, freq=split_delta))
    next(e_ranges, None)  # skip to the next item in the 2nd iterable
    # Zip them together to get a list of (start, end) tuples of ranges
    # Note: we subtract 1 nanosecond from the 'end' value of each range so
    # to avoid getting duplicated records at the boundaries of the ranges.
    # Some providers don't have nanosecond granularity so we might
    # get duplicates in these cases
    ranges: list[tuple[datetime, datetime]] = [
        (s_time, e_time - pd.Timedelta("1ns"))
        for s_time, e_time in zip(s_ranges, e_ranges)
    ]

    # Since the generated time ranges are based on deltas from 'start'
    # we need to adjust the end time on the final range.
    # If the difference between the calculated last range end and
    # the query 'end' that the user requested is small (< 0.1% of a delta),
    # we just replace the last "end" time with our query end time.
    if (end - ranges[-1][1]) < (split_delta / 1000):
        ranges[-1] = ranges[-1][0], pd.Timestamp(end)
    else:
        # otherwise append a new range starting after the last range
        # in ranges and ending in 'end"
        # note - we need to add back our subtracted 1 nanosecond
        ranges.append((ranges[-1][1] + pd.Timedelta("1ns"), pd.Timestamp(end)))

    return ranges



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
import asyncio
from typing import Any, Iterable, Optional

import ipywidgets as widgets
from IPython.display import display

from .._version import VERSION
from .core import IPyDisplayMixin

__version__ = VERSION
__author__ = "Ian Hellen"


class OptionButtons(IPyDisplayMixin):
    """
    OptionButtons creates a sequence of buttons to choose from.

    The widget can be run in synchronous mode as a simple option
    selector or in async mode with a timeout.
    In the latter mode, after the timeout has expired the widget
    value is set to the default option button value.
    To use the async mode you must call `widget.display_async()` with
    the async keyword.

    Attributes
    ----------
    value : str
        The value of the option selected (case-normalized)

    Example
    -------
    >>> opt = OptionButtons(description="Continue something?",
    ...  buttons=["Maybe", "Yes", "Cancel"], timeout=10)
    >>> await opt.display_async()

    """

    def __init__(
        self,
        description: Optional[str] = "Select an option to continue",
        buttons: Optional[Iterable[str]] = None,
        default: Optional[str] = None,
        timeout: int = 0,
        debug: bool = False,
        **kwargs,
    ):
        """
        Initialize the OptionButton widget.

        Parameters
        ----------
        description : Optional[str], optional
            Description label displayed above the buttons,
            by default "Select an option to continue"
        buttons : Optional[Iterable[str]], optional
            A list of button values, by default None. This
            will default to ["Yes", "No", "Cancel"]
            "options" is an alias for this parameter.
        default : Optional[str], optional
            The default value to use on timeout, by default the
            first value in the `buttons` list
        timeout : int, optional
            Timeout in seconds, by default 0
        debug : bool, optional
            Adds some debug information to an Output controle,
            by default False

        """
        buttons = buttons or kwargs.pop("options", None)
        if buttons is None:
            buttons = ["Yes", "No", "Cancel"]
        self._buttons = [widgets.Button(description=b_item) for b_item in buttons]
        self._desc_label = widgets.Label(value=description)
        self._timer_label = widgets.Label(layout=widgets.Layout(left="10px"))
        self.default = default or next(iter(buttons)).casefold()
        self.value: Optional[str] = None
        self.timeout = timeout

        self._completion: Any = None
        self._fut_val: Any = None
        self._debug = debug
        if self._debug:
            self._out = widgets.Output()
        self._create_button_callbacks(self._buttons)

    @property
    def layout(self):
        """Create layout for buttons."""
        return widgets.VBox(
            [self._desc_label, widgets.HBox([*(self._buttons), self._timer_label])]
        )

    def _debug_out(self, mssg: str):
        if self._debug:
            self._out.append_stdout(mssg)

    def _create_button_callbacks(self, buttons):
        """Set up buttons."""

        def getvalue(change):
            """Button on_click handler."""
            self.value = change.description
            for btn in buttons:
                btn.on_click(getvalue, remove=True)

        for btn in buttons:
            btn.on_click(getvalue)

    async def _await_widget(self):
        """Awaitable coroutine for widget."""
        self._debug_out("await_widget entered\n")
        self._create_button_callbacks(self._buttons)
        self._debug_out("buttons set\n")

        done, _ = await asyncio.wait(
            [self._wait_for_button_change(), self._await_timer(self.timeout)],
            return_when=asyncio.FIRST_COMPLETED,
            timeout=self.timeout + 5,
        )
        self._debug_out("wait returned\n")
        self._completion = done
        self._debug_out(str(done))
        return done

    async def _wait_for_button_change(self):
        """Awaitable for button selection state."""
        self._debug_out("wait_for_button_change entered\n")
        while self.value is None:
            await asyncio.sleep(0.1)
            if self._debug:
                self._debug_out("*")

    async def _await_timer(self, timeout: int = 5):
        timeout = max(timeout, 0)
        while timeout > 0:
            self._timer_label.value = f"Waiting {timeout} sec..."
            if self.value:
                self._timer_label.value = f"Option selected: '{self.value}'"
                return
            await asyncio.sleep(1)
            timeout -= 1
        self.value = self.default
        self._timer_label.value = f"Timed out. Defaulted to '{self.value}'"

    async def display_async(self, reset: bool = False):
        """
        Display widget with timeout.

        Parameters
        ----------
        reset : bool, optional
            Resets any current value to None,
            by default False

        """
        if reset:
            self.value = None
        display(self.layout)
        if self._debug:
            display(self._out)
        self._fut_val = asyncio.ensure_future(self._await_widget())
        self._debug_out("future returned\n")
        self._debug_out(str(self._fut_val) + "\n")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Formatting and checking functions."""
import builtins
import re
import uuid
from typing import Any

from ..._version import VERSION
from .types import export

__version__ = VERSION
__author__ = "Ian Hellen"


@export
def is_valid_uuid(uuid_str: str) -> bool:
    """
    Return true if `uuid_str` is a value GUID/UUID.

    Parameters
    ----------
    uuid_str : str
        String to test

    Returns
    -------
    bool
        True if valid GUID/UUID.

    """
    if not uuid_str:
        return False
    try:
        uuid.UUID(uuid_str)
    except (ValueError, TypeError):
        return False
    return True


@export
def valid_pyname(identifier: str) -> str:
    """
    Return legal Python identifier, which doesn't collide with builtins.

    Parameters
    ----------
    identifier : str
        The input identifier

    Returns
    -------
    str
        The cleaned identifier

    """
    builtin_names = set(dir(builtins))
    if identifier in builtin_names:
        identifier = f"{identifier}_bi"
    identifier = re.sub("[^a-zA-Z0-9_]", "_", identifier)
    if identifier[0].isdigit():
        identifier = f"n_{identifier}"
    return identifier


@export
def string_empty(string: str) -> bool:
    """Return True if the input string is None or whitespace."""
    return not (bool(string) or isinstance(string, str) and bool(string.strip()))


@export
def is_not_empty(test_object: Any) -> bool:
    """Return True if the test_object is not None or empty."""
    if test_object:
        return bool(test_object.strip()) if isinstance(test_object, str) else True
    return False


# String escapes
@export
def escape_windows_path(str_path: str) -> str:
    """Escape backslash characters in a string."""
    return str_path.replace("\\", "\\\\") if is_not_empty(str_path) else str_path


@export
def unescape_windows_path(str_path: str) -> str:
    """Remove escaping from backslash characters in a string."""
    return str_path.replace("\\\\", "\\") if is_not_empty(str_path) else str_path


@export
def defang_ioc(ioc: str, ioc_type: str = None) -> str:
    """
    Return de-fanged observable.

    Parameters
    ----------
    ioc : str
        The observable.
    ioc_type : str
        The type of IoC. If URL or Email it will do
        extra processing to neuter the URL protocol and email @ symbol

    Returns
    -------
    str
        The de-fanged observable.

    """
    de_fanged = ioc
    if ioc_type == "email":
        de_fanged = de_fanged.replace("@", "[AT]")
    elif ioc_type == "url":
        de_fanged = de_fanged.replace("http", "hXXp").replace("ftp", "fXp")
    return de_fanged.replace(".", "[.]")


@export
def refang_ioc(ioc: str, ioc_type: str = None) -> str:
    """
    Return observable with removed de-fanging elements.

    Parameters
    ----------
    ioc : str
        The observable to re-fang.
    ioc_type : str
        The type of IoC. If URL or Email it will do
        extra processing to restore protocol and @ symbol

    Returns
    -------
    str
        The re-fanged observable.

    """
    re_fanged = ioc
    if ioc_type == "email":
        re_fanged = re_fanged.replace("[AT]", "@")
    elif ioc_type == "url":
        re_fanged = re_fanged.replace("hXXp", "http").replace("fXp", "ftp")
    return re_fanged.replace("[.]", ".")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""msticpy IPython magics."""
import re
from typing import List, Tuple

# pylint: enable=unused-import
from IPython import get_ipython
from IPython.core import magic_arguments
from IPython.core.magic import (
    Magics,
    line_cell_magic,
    magics_class,
    needs_local_scope,
    register_line_cell_magic,
    register_line_magic,
)

from ..init.pivot_core.pivot_magic_core import run_txt2df

try:
    from bs4 import BeautifulSoup

    _BS_AVAILABLE = True
except ImportError:
    _BS_AVAILABLE = False

from .._version import VERSION
from ..common.utility import is_ipython
from ..common.utility.ipython import save_obj_to_cell
from ..transform import base64unpack as base64
from ..transform.iocextract import IoCExtract

__version__ = VERSION
__author__ = "Ian Hellen"


if is_ipython():

    @magics_class
    class Base64Magic(Magics):
        """Base64 IPython magic extension."""

        _STRIP_TAGS = r"</?decoded[^>]*>"

        @line_cell_magic
        @magic_arguments.magic_arguments()
        @magic_arguments.argument(
            "--out", "-o", help="The variable to return the results in"
        )
        @magic_arguments.argument(
            "--pretty",
            "-p",
            help="Print formatted version of output",
            action="store_true",
        )
        @magic_arguments.argument(
            "--clean",
            "-c",
            help="Print decoded string with no formatting",
            action="store_true",
        )
        @magic_arguments.argument(
            "--utf16",
            "-u",
            help="Decode to UTF-16.",
            default=False,
            action="store_true",
        )
        def b64(self, line: str = "", cell: str = None) -> str:
            """
            Base64 IPython magic extension.

            Parameters
            ----------
            line : str, optional
                Line contents, by default ""
            cell : str, optional
                Cell contents, by default None

            Returns
            -------
            str
                Decoded text

            """
            if cell is None:
                results, _ = base64.unpack(line)
                return results

            args = magic_arguments.parse_argstring(self.b64, line)
            results, df_results = base64.unpack(cell, utf16=args.utf16)

            if args.clean:
                results = re.sub(self._STRIP_TAGS, "", results)
            elif args.pretty:
                if _BS_AVAILABLE:
                    xml_str = f"<decoded_string>{results}</decoded_string>"
                    b_soup = BeautifulSoup(xml_str, "xml")
                    results = b_soup.prettify()
            if args.out is not None:
                self.shell.user_ns[args.out] = (results, df_results)  # type: ignore
            return results

    @magics_class
    class IoCExtractMagic(Magics):
        """Ioc Extract IPython magic extension."""

        def __init__(self, shell):
            """
            Instantiate magic class.

            Parameters
            ----------
            shell : IPython shell
                IPython shell

            """
            # You must call the parent constructor
            super().__init__(shell)
            self._ioc_extract = IoCExtract()

        @line_cell_magic
        @magic_arguments.magic_arguments()
        @magic_arguments.argument(
            "--out", "-o", help="The variable to return the results in"
        )
        @magic_arguments.argument(
            "--ioc_types",
            "-i",
            help="The types of IoC to search for (comma-separated string)",
        )
        def ioc(self, line="", cell=None) -> List[Tuple[str, List[str]]]:
            """
            Ioc Extract IPython magic extension.

            Parameters
            ----------
            line : str, optional
                Line contents, by default ""
            cell : str, optional
                Cell contents, by default None

            Returns
            -------
            List[Tuple[str, List[str]]]
                List of tuples of IoCs found grouped by type.

            """
            args = magic_arguments.parse_argstring(self.ioc, line)
            ioc_types = None
            if args.ioc_types:
                ioc_types = [ioc_type.strip() for ioc_type in args.ioc_types.split(",")]

            if cell is None:
                results = self._ioc_extract.extract(src=line, ioc_types=ioc_types)
            else:
                results = self._ioc_extract.extract(src=cell, ioc_types=ioc_types)
            iocs = [(ioc_type, list(ioc_res)) for ioc_type, ioc_res in results.items()]

            if args.out is not None:
                self.shell.user_ns[args.out] = results  # type: ignore
            return iocs

    IPYTHON = get_ipython()
    if IPYTHON:
        IPYTHON.register_magics(Base64Magic)
        IPYTHON.register_magics(IoCExtractMagic)

    @register_line_cell_magic
    @needs_local_scope
    def txt2df(line, cell, local_ns):  # noqa: D417
        """
        Convert cell text to pandas DataFrame.

        Parameters
        ----------
        sep : str
            Column separator/delimiter to use
            `--sep ;` or `-s ,`, default ",",
        name : str
            If specified, the DataFrame will be assigned to the named variable.
            `--name df_name`" or `-n df_name`,
        headers : bool
            If supplied, the first line is treated as the header row.
            `--headers` or `-h`
        keepna : bool
            Don't drop columns that are all NA, The default is to drop
            them, which is useful for data with trailing delimiters)
            `--keepna` or `-k`

        """
        return run_txt2df(line, cell, local_ns)

    @register_line_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument(
        "--var_name",
        "-v",
        help="The variable to return the results in.",
    )
    @magic_arguments.argument(
        "--object",
        "-o",
        help="The object name to store.",
    )
    @needs_local_scope
    def save_to_cell(line: str, local_ns: dict):
        """
        Msticpy SaveToCell IPython magic extension.

        Parameters
        ----------
        line : str, optional
            Line contents, by default ""

        """
        args = magic_arguments.parse_argstring(save_to_cell, line)
        print(args)
        print(type(args.object))
        obj = local_ns.get(args.object)
        if obj is None:
            raise ValueError(f"Variable '{args.object}' not found.")
        save_obj_to_cell(obj, args.var_name)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for common display functions."""
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import networkx as nx
from bokeh.io import output_notebook
from bokeh.models import (  # type: ignore[attr-defined]
    BoxSelectTool,
    Circle,
    EdgesAndLinkedNodes,
    HoverTool,
    Label,
    MultiLine,
    NodesAndLinkedEdges,
    Renderer,
    TapTool,
    WheelZoomTool,
)
from bokeh.palettes import Spectral4
from bokeh.plotting import figure, from_networkx, show
from typing_extensions import Literal

from .._version import VERSION
from .figure_dimension import bokeh_figure

__version__ = VERSION
__author__ = "Ian Hellen"

# wrap figure function to handle v2/v3 parameter renaming
figure = bokeh_figure(figure)  # type: ignore[assignment, misc]


GraphLayout = Union[
    Callable[[Any], Dict[str, Tuple[float, float]]],
    Literal[
        "spring",
        "bipartite",
        "circular",
        "kamada_kawai",
        "planar",
        "random",
        "shell",
        "spectral",
        "spiral",
        "multi_partite",
    ],
    Dict[str, Tuple[float, float]],
]


# pylint: disable=too-many-arguments, too-many-locals
def plot_nx_graph(
    nx_graph: nx.Graph,
    title: str = "Data Graph",
    node_size: int = 25,
    font_size: Union[int, str] = 10,
    height: int = 800,
    width: int = 800,
    scale: int = 2,
    hide: bool = False,
    source_attrs: Optional[Iterable[str]] = None,
    target_attrs: Optional[Iterable[str]] = None,
    edge_attrs: Optional[Iterable[str]] = None,
    layout: GraphLayout = "spring",
    **kwargs,
) -> figure:
    """
    Plot entity graph with Bokeh.

    Parameters
    ----------
    nx_graph : nx.Graph
        The entity graph as a networkX graph
    title : str
        Title for the plot, by default 'Data Graph'
    node_size : int, optional
        Size of the nodes in pixels, by default 25
    font_size : int, optional
        Font size for node labels, by default 10
        Can be an integer (point size) or a string (e.g. "10pt")
    width : int, optional
        Width in pixels, by default 800
    height : int, optional
        Image height (the default is 800)
    scale : int, optional
        Position scale (the default is 2)
    hide : bool, optional
        Don't show the plot, by default False. If True, just
        return the figure.
    source_attrs : Optional[List[str]], optional
        Optional list of source attributes to use as hover properties, by default None
    target_attrs : Optional[List[str]], optional
        Optional list of target attributes to use as hover properties, by default None
    edge_attrs : Optional[List[str]], optional
        Optional list of edge attributes to use as hover properties, by default None
    layout : GraphLayout, optional
        The layout name to use, a callable that will return layout
        positions, a dictionary of {nFode_id: (float, float)} node positions.
        The default is "spring".
        See https://networkx.org/documentation/stable/reference/drawing.html

    Other Parameters
    ----------------
    source_color : str, optional
        The color of the source nodes, by default 'light-blue'
    target_color : str, optional
        The color of the source nodes, by default 'light-green'
    edge_color : str, optional
        The color of the edges, by default 'black'
    kwargs :
        Additional keyword arguments are passed to the networkx
        layout function.

    Returns
    -------
    bokeh.plotting.figure
        The network plot.

    """
    output_notebook()
    font_pnt = f"{font_size}pt" if isinstance(font_size, int) else font_size
    source_color = kwargs.pop("source_color", "lightblue")
    target_color = kwargs.pop("target_color", "lightgreen")
    edge_color = kwargs.pop("edge_color", "black")
    node_attrs = {
        node: attrs.get(
            "color",
            (
                source_color
                if attrs.get("node_role", "source") == "source"
                else target_color
            ),
        )
        for node, attrs in nx_graph.nodes(data=True)
    }
    nx.set_node_attributes(nx_graph, node_attrs, "node_color")
    hide = kwargs.pop("hide", False)

    plot = figure(
        title=title,
        x_range=(-3, 3),
        y_range=(-3, 3),
        width=width,
        height=height,
    )

    nx_graph_for_plotting = nx.Graph()
    index_node = 0
    rev_index = {}
    fwd_index = {}
    node_attributes = {}
    for node_key in nx_graph.nodes:
        nx_graph_for_plotting.add_node(index_node)
        rev_index[node_key] = index_node
        fwd_index[index_node] = node_key
        node_attributes[index_node] = nx_graph.nodes[node_key]
        index_node += 1

    nx.set_node_attributes(nx_graph_for_plotting, node_attributes)

    for source_node, target_node in nx_graph.edges:
        nx_graph_for_plotting.add_edge(rev_index[source_node], rev_index[target_node])

    graph_layout = _get_graph_layout(nx_graph_for_plotting, layout, **kwargs)

    graph_renderer = from_networkx(
        nx_graph_for_plotting, graph_layout, scale=scale, center=(0, 0)
    )
    _create_edge_renderer(graph_renderer, edge_color=edge_color)
    _create_node_renderer(graph_renderer, node_size, "node_color")

    graph_renderer.selection_policy = NodesAndLinkedEdges()
    graph_renderer.inspection_policy = EdgesAndLinkedNodes()
    # pylint: disable=no-member
    plot.renderers.append(graph_renderer)  # type: ignore[attr-defined]

    hover_tools = [
        _create_node_hover(source_attrs, target_attrs, [graph_renderer.node_renderer])
    ]
    if edge_attrs:
        hover_tools.append(
            _create_edge_hover(edge_attrs, [graph_renderer.edge_renderer])
        )
    plot.add_tools(*hover_tools, WheelZoomTool(), TapTool(), BoxSelectTool())

    # Create labels
    # pylint: disable=no-member
    for index, pos in graph_renderer.layout_provider.graph_layout.items():
        label = Label(
            x=pos[0],
            y=pos[1],
            x_offset=5,
            y_offset=5,
            text=fwd_index[int(index)],
            text_font_size=font_pnt,
        )
        plot.add_layout(label)
    if not hide:
        show(plot)
    return plot


def _get_graph_layout(nx_graph: nx.Graph, layout: GraphLayout, **kwargs):
    """Return a layout for the graph."""
    if callable(layout):
        return layout(nx_graph, **kwargs)
    layout_func = getattr(nx, f"{layout}_layout", None)
    if layout_func and callable(layout_func):
        # pylint: disable=not-callable
        return layout_func(nx_graph, **kwargs)
    return nx.spring_layout(nx_graph, **kwargs)


def _create_node_hover(
    source_attrs: Optional[Iterable[str]],
    target_attrs: Optional[Iterable[str]],
    renderers: List[Renderer],
) -> HoverTool:
    """Create a hover tool for nodes."""
    node_attr_cols = set((list(source_attrs or [])) + (list(target_attrs or [])))
    node_tooltips = [
        ("node_type", "@node_type"),
        *[(col, f"@{{{col}}}") for col in node_attr_cols],
    ]
    return HoverTool(tooltips=node_tooltips, renderers=renderers)


def _create_edge_hover(
    edge_attrs: Iterable[str], renderers: List[Renderer]
) -> HoverTool:
    """Create a hover tool for nodes."""
    edge_attr_cols = edge_attrs or []
    edge_tooltips = [
        *[(col, f"@{{{col}}}") for col in edge_attr_cols],
    ]
    return HoverTool(tooltips=edge_tooltips, renderers=renderers)


def _create_node_renderer(graph_renderer: Renderer, node_size: int, fill_color: str):
    """Create graph render for nodes."""
    graph_renderer.node_renderer.glyph = Circle(size=node_size, fill_color=fill_color)
    graph_renderer.node_renderer.hover_glyph = Circle(
        size=node_size, fill_color=Spectral4[1]
    )
    graph_renderer.node_renderer.selection_glyph = Circle(
        size=node_size, fill_color=Spectral4[2]
    )


def _create_edge_renderer(graph_renderer: Renderer, edge_color: str):
    """Create graph render for edges."""
    graph_renderer.edge_renderer.hover_glyph = MultiLine(
        line_color=Spectral4[1], line_width=5
    )
    graph_renderer.edge_renderer.glyph = MultiLine(
        line_alpha=0.8, line_color=edge_color, line_width=1
    )
    graph_renderer.edge_renderer.selection_glyph = MultiLine(
        line_color=Spectral4[2], line_width=5
    )


def plot_entity_graph(
    entity_graph: nx.Graph,
    node_size: int = 25,
    font_size: Union[int, str] = 10,
    height: int = 800,
    width: int = 800,
    scale: int = 2,
    hide: bool = False,
) -> figure:
    """
    Plot entity graph with Bokeh.

    Parameters
    ----------
    entity_graph : nx.Graph
        The entity graph as a networkX graph
    node_size : int, optional
        Size of the nodes in pixels, by default 25
    font_size : int, optional
        Font size for node labels, by default 10
        Can be an integer (point size) or a string (e.g. "10pt")
    width : int, optional
        Width in pixels, by default 800
    height : int, optional
        Image height (the default is 800)
    scale : int, optional
        Position scale (the default is 2)
    hide : bool, optional
        Don't show the plot, by default False. If True, just
        return the figure.

    Returns
    -------
    bokeh.plotting.figure
        The network plot.

    """
    output_notebook()
    font_pnt = f"{font_size}pt" if isinstance(font_size, int) else font_size
    node_attrs = {
        node: attrs.get("color", "green")
        for node, attrs in entity_graph.nodes(data=True)
    }
    nx.set_node_attributes(entity_graph, node_attrs, "node_color")

    plot = figure(
        title="Alert Entity graph",
        x_range=(-3, 3),
        y_range=(-3, 3),
        width=width,
        height=height,
    )

    plot.add_tools(
        HoverTool(
            tooltips=[
                ("node_type", "@node_type"),
                ("name", "@name"),
                ("description", "@description"),
                ("entitytype", "@entitytype"),
            ]
        )
    )

    graph_renderer = from_networkx(
        entity_graph, nx.spring_layout, scale=scale, center=(0, 0)
    )
    graph_renderer.node_renderer.glyph = Circle(
        size=node_size, fill_color="node_color", fill_alpha=0.5
    )
    # pylint: disable=no-member
    plot.renderers.append(graph_renderer)  # type: ignore[attr-defined]

    # Create labels
    for name, pos in graph_renderer.layout_provider.graph_layout.items():
        label = Label(
            x=pos[0],
            y=pos[1],
            x_offset=5,
            y_offset=5,
            text=str(name),
            text_font_size=font_pnt,
        )
        plot.add_layout(label)
    # pylint: enable=no-member
    if not hide:
        show(plot)
    return plot



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Class for Sentinel Workspaces."""
from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar
from urllib import parse

import httpx
from msrestazure import tools as az_tools
from typing_extensions import Self

from msticpy.context.azure.sentinel_utils import SentinelUtilsMixin

from ..._version import VERSION
from ...auth.azure_auth_core import AzureCloudConfig
from ...common.data_utils import df_has_data
from ...common.pkg_config import get_http_timeout
from ...common.utility import mp_ua_header
from ...data.core.data_providers import QueryProvider

if TYPE_CHECKING:
    import pandas as pd

logger: logging.Logger = logging.getLogger(__name__)

__version__ = VERSION
__author__ = "Ian Hellen"


@dataclass
class ParsedUrlComponents:
    """Class to defined components for Parsed URLs."""

    domain: str | None
    resource_id: str
    tenant_name: str | None
    res_components: dict[str, str]
    raw_res_id: str


class SentinelWorkspacesMixin(SentinelUtilsMixin):
    """Mixin class for Sentinel workspaces."""

    _TENANT_URI: ClassVar[str] = (
        "{cloud_endpoint}/{tenant_name}/.well-known/openid-configuration"
    )
    _RES_GRAPH_PROV: ClassVar[QueryProvider | None] = None

    @classmethod
    def get_resource_id_from_url(
        cls: type[Self],
        portal_url: str,
    ) -> str | None:
        """Return resource ID components from Sentinel portal URL."""
        if (resource := cls._extract_resource_id(portal_url)) is not None:
            return resource.resource_id
        return None

    @classmethod
    def get_workspace_details_from_url(
        cls: type[Self],
        portal_url: str,
    ) -> dict[str, dict[str, str]]:
        """
        Return workspace settings from portal URL.

        Parameters
        ----------
        portal_url : str
            URL from Sentinel Azure portal

        Returns
        -------
        dict[str, dict[str, str]]

        """
        resource_comps: ParsedUrlComponents | None = cls._extract_resource_id(
            portal_url,
        )
        if not resource_comps:
            err_msg: str = f"Cannot retrieve workspace details from {portal_url}"
            raise ValueError(err_msg)
        tenant_id: str | None = None
        if resource_comps.tenant_name:
            tenant_id = cls._get_tenantid_from_logon_domain(resource_comps.tenant_name)
        workspace_df: pd.DataFrame = cls._lookup_workspace_by_res_id(
            resource_id=resource_comps.resource_id,
        )
        if df_has_data(workspace_df):
            return cls._get_settings_for_workspace(
                workspace_name=workspace_df.iloc[0].workspaceName,
                workspace_id=workspace_df.iloc[0].workspaceId,
                tenant_id=tenant_id or workspace_df.iloc[0].tenantId,
                subscription_id=workspace_df.iloc[0].subscriptionId,
                resource_group=workspace_df.iloc[0].resourceGroup,
                workspace_tenant_id=workspace_df.iloc[0].tenantId,
            )
        logger.warning(
            "Failed to find Azure resource for workspace. Returning partial results.",
        )
        return cls._get_settings_for_workspace(
            workspace_name=resource_comps.res_components["name"],
            workspace_id="unknown",
            tenant_id=tenant_id or "unknown",
            subscription_id=resource_comps.res_components["subscription"],
            resource_group=resource_comps.res_components["resource_group"],
            workspace_tenant_id="unknown",
        )

    @classmethod
    def get_workspace_name(
        cls: type[Self],
        workspace_id: str | None = None,
        resource_id: str | None = None,
    ) -> str | None:
        """
        Return resolved name from workspace ID or resource ID.

        Parameters
        ----------
        workspace_id : Optional[str], optional
            The UUID of the Sentinel workspace, by default None
        resource_id : Optional[str], optional
            The Resource ID string of the workspace, by default None

        Returns
        -------
        Optional[str]
            The workspace name, if found, else None

        Raises
        ------
        ValueError
            If neither workspace_id or resource_id parameters are supplied.

        """
        settings: dict[str, Any] = cls.get_workspace_settings(
            workspace_id=workspace_id,
            resource_id=resource_id,
        )

        return next(iter(settings.values())).get("WorkspaceName") if settings else None

    @classmethod
    def get_workspace_id(
        cls: type[Self],
        workspace_name: str,
        subscription_id: str = "",
        resource_group: str = "",
    ) -> str | None:
        """
        Return the workspace ID given workspace name.

        Parameters
        ----------
        workspace_name : str
            Workspace name (case insensitive)
        subscription_id : str, optional
            Azure subscription UUID, by default ""
        resource_group : str, optional
            Azure resource group name, by default ""

        Returns
        -------
        Optional[str]
            The ID of the workspace if found, else None

        """
        settings: dict[str, Any] = cls.get_workspace_settings_by_name(
            workspace_name,
            subscription_id,
            resource_group,
        )
        return next(iter(settings.values())).get("WorkspaceId") if settings else None

    @classmethod
    def get_workspace_settings(
        cls: type[Self],
        workspace_id: str | None = None,
        resource_id: str | None = None,
    ) -> dict[str, Any]:
        """
        Return resolved workspace settings from workspace ID or resource ID.

        Parameters
        ----------
        workspace_id : Optional[str], optional
            The UUID of the Sentinel workspace, by default None
        resource_id : Optional[str], optional
            The Resource ID string of the workspace, by default None

        Returns
        -------
        dict[str, str]
            The workspace name, if found, else None

        Raises
        ------
        ValueError
            If neither workspace_id or resource_id parameters are supplied.

        """
        if workspace_id:
            results_df: pd.DataFrame = cls._lookup_workspace_by_ws_id(workspace_id)
        else:
            if not resource_id:
                err_msg: str = "Either workspace_id or resource_id must be supplied."
                raise ValueError(err_msg)
            results_df = cls._lookup_workspace_by_res_id(resource_id)
        if df_has_data(results_df):
            return cls._get_settings_for_workspace(
                workspace_name=results_df.iloc[0].workspaceName,
                workspace_id=results_df.iloc[0].workspaceId,
                tenant_id=results_df.iloc[0].tenantId,
                subscription_id=results_df.iloc[0].subscriptionId,
                resource_group=results_df.iloc[0].resourceGroup,
                workspace_tenant_id=results_df.iloc[0].tenantId,
            )
        return {}

    @classmethod
    def get_workspace_settings_by_name(
        cls: type[Self],
        workspace_name: str,
        subscription_id: str = "",
        resource_group: str = "",
    ) -> dict[str, Any]:
        """
        Return the workspace ID given workspace name.

        Parameters
        ----------
        workspace_name : str
            Workspace name (case insensitive)
        subscription_id : str, optional
            Azure subscription UUID, by default ""
        resource_group : str, optional
            Azure resource group name, by default ""

        Returns
        -------
        Optional[str]
            The ID of the workspace if found, else None

        """
        results_df: pd.DataFrame = cls._lookup_workspace_by_name(
            workspace_name,
            subscription_id,
            resource_group,
        )
        if df_has_data(results_df):
            if len(results_df) > 1:
                logger.warning(
                    "Warning: query returned multiple results. "
                    "Specify subscription_id and/or resource_group "
                    "for more accurate results.",
                )
            return cls._get_settings_for_workspace(
                workspace_name=results_df.iloc[0].workspaceName,
                workspace_id=results_df.iloc[0].workspaceId,
                tenant_id=results_df.iloc[0].tenantId,
                subscription_id=results_df.iloc[0].subscriptionId,
                resource_group=results_df.iloc[0].resourceGroup,
                workspace_tenant_id=results_df.iloc[0].tenantId,
            )
        return {}

    @classmethod
    def _get_resource_graph_provider(
        cls: type[Self],
    ) -> QueryProvider:
        if not cls._RES_GRAPH_PROV:
            cls._RES_GRAPH_PROV = QueryProvider("ResourceGraph")
        if not cls._RES_GRAPH_PROV.connected:
            cls._RES_GRAPH_PROV.connect()  # pragma: no cover
        return cls._RES_GRAPH_PROV

    @classmethod
    def _lookup_workspace_by_name(
        cls: type[Self],
        workspace_name: str,
        subscription_id: str = "",
        resource_group: str = "",
    ) -> pd.DataFrame:
        res_graph_prov: QueryProvider = cls._get_resource_graph_provider()
        return res_graph_prov.Sentinel.list_sentinel_workspaces_for_name(
            workspace_name=workspace_name,
            subscription_id=subscription_id,
            resource_group=resource_group,
        )

    @classmethod
    def _lookup_workspace_by_ws_id(
        cls: type[Self],
        workspace_id: str,
    ) -> pd.DataFrame:
        res_graph_prov: QueryProvider = cls._get_resource_graph_provider()
        return res_graph_prov.Sentinel.get_sentinel_workspace_for_workspace_id(
            workspace_id=workspace_id,
        )

    @classmethod
    def _lookup_workspace_by_res_id(
        cls: type[Self],
        resource_id: str | None,
    ) -> pd.DataFrame:
        res_graph_prov: QueryProvider = cls._get_resource_graph_provider()
        return res_graph_prov.Sentinel.get_sentinel_workspace_for_resource_id(
            resource_id=resource_id,
        )

    @classmethod
    def _extract_resource_id(
        cls: type[Self],
        url: str,
    ) -> ParsedUrlComponents | None:
        """Extract and return resource ID components from URL."""
        resid_pattern = (
            r"https://(?P<domain>[^/]+)/#?(@(?P<tenantname>[^/]+))?"
            ".*(?P<res_id>(%2F|/)subscriptions(%2F|/).*)"
        )

        uri_match: re.Match[str] | None = re.search(resid_pattern, url)
        if not uri_match:
            return None

        raw_res_id: str = uri_match.groupdict()["res_id"]
        raw_res_id = parse.unquote(raw_res_id)
        res_components: dict[str, Any] = az_tools.parse_resource_id(raw_res_id)
        try:
            resource_id: str = cls._normalize_resource_id(res_components)
        except KeyError:
            logger.exception("Invalid Sentinel resource id")
            return None
        return ParsedUrlComponents(
            domain=uri_match.groupdict().get("domain"),
            resource_id=resource_id,
            tenant_name=uri_match.groupdict().get("tenantname"),
            res_components=res_components,
            raw_res_id=raw_res_id,
        )

    @staticmethod
    def _normalize_resource_id(res_components: dict[str, str]) -> str:
        return (
            f"/subscriptions/{res_components['subscription']}"
            f"/resourcegroups/{res_components['resource_group']}"
            f"/providers/Microsoft.OperationalInsights"
            f"/workspaces/{res_components['name']}"
        )

    @classmethod
    def _get_tenantid_from_logon_domain(
        cls: type[Self],
        domain: str,
        cloud: str = "global",
    ) -> str | None:
        """Get the tenant ID from login domain."""
        az_cloud_config = AzureCloudConfig(cloud)
        login_endpoint: str = az_cloud_config.authority_uri
        t_resp: httpx.Response = httpx.get(
            cls._TENANT_URI.format(cloud_endpoint=login_endpoint, tenant_name=domain),
            timeout=get_http_timeout(),
            headers=mp_ua_header(),
        )
        tenant_details: dict[str, Any] = t_resp.json()

        if not tenant_details:
            return None
        tenant_ep_rgx = r"(?P<endpoint>https://[^/]+)/(?P<tenant_id>[^/]+).*"
        match: re.Match[str] | None = re.search(
            tenant_ep_rgx,
            tenant_details.get("token_endpoint", ""),
        )
        return match.groupdict()["tenant_id"] if match else None

    @classmethod
    def _get_settings_for_workspace(  # pylint:disable=too-many-arguments # noqa:PLR0913
        cls: type[Self],
        workspace_name: str,
        workspace_id: str,
        tenant_id: str,
        subscription_id: str,
        resource_group: str,
        workspace_tenant_id: str,
    ) -> dict[str, dict[str, str]]:
        """Return settings dictionary for workspace settings."""
        return {
            workspace_name: {
                "WorkspaceId": workspace_id,
                "TenantId": tenant_id,
                "SubscriptionId": subscription_id,
                "ResourceGroup": resource_group,
                "WorkspaceName": workspace_name,
                "WorkspaceTenantId": workspace_tenant_id,
            },
        }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTICPy Data Processing Tools."""

from .._version import VERSION
from ..lazy_importer import lazy_import

__version__ = VERSION

_LAZY_IMPORTS = {
    "msticpy.context.geoip.GeoLiteLookup",
    "msticpy.context.geoip.IPStackLookup",
    "msticpy.context.geoip.geo_distance",
    "msticpy.context.tilookup.TILookup",
    "msticpy.transform.iocextract.IoCExtract",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Result Severity enumeration."""
from __future__ import annotations

from enum import Enum
from functools import total_ordering
from typing import Any, NamedTuple

from typing_extensions import Self

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@total_ordering
class ResultSeverity(Enum):
    """Item report severity."""

    # pylint: disable=invalid-name
    unknown = -1
    information = 0
    warning = 1
    high = 2

    # pylint: enable=invalid-name

    @classmethod
    def parse(cls: type[Self], value: object) -> ResultSeverity:
        """
        Parse string or numeric value to ResultSeverity.

        Parameters
        ----------
        value : Any
            ResultSeverity, str or int

        Returns
        -------
        ResultSeverity
            ResultSeverity instance.

        """
        if isinstance(value, ResultSeverity):
            return value
        if isinstance(value, str) and value.lower() in cls.__members__:
            return cls[value.lower()]
        if isinstance(value, int) and value in [
            v.value for v in cls.__members__.values()
        ]:
            return cls(value)
        return ResultSeverity.unknown

    def __eq__(self: Self, other: object) -> bool:
        """
        Return True if severities are equal.

        Parameters
        ----------
        other : Any
            ResultSeverity to compare to.
            Can be a numeric value or name of ResultSeverity value.

        Returns
        -------
        bool
            If severities are equal

        """
        other_sev: ResultSeverity = ResultSeverity.parse(other)
        return self.value == other_sev.value

    def __gt__(self: Self, other: str | int | ResultSeverity) -> bool:
        """
        Return True self is greater than other.

        Parameters
        ----------
        other : Any
            ResultSeverity to compare to.
            Can be a numeric value or name of ResultSeverity value.

        Returns
        -------
        bool
            If severities are equal

        """
        other_sev: ResultSeverity = ResultSeverity.parse(other)
        return self.value > other_sev.value


class LookupResult(NamedTuple):
    """Return value from TI Provider parse_results."""

    status: bool
    severity: ResultSeverity
    details: dict[str, Any]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPy User Session loader.

This loads user-specific configuration settings from a YAML file.

Example YAML file:

.. code-block:: yaml

    QueryProviders:
        qry_prov_sent:
            DataEnvironment: MSSentinel
            InitArgs:
                debug: True
            Connect: True
            ConnectArgs:
                workspace: MySoc
                auth_methods: ['cli', 'device_code']
        qry_prov_md:
            DataEnvironment: M365D
        qry_kusto_mde:
            DataEnvironment: Kusto
            Connect: True
            ConnectArgs:
                cluster: MDEData
        qry_kusto_mstic:
            DataEnvironment: Kusto
            Connect: True
            ConnectArgs:
                cluster: MSTIC
    Components:
        mssentinel:
            Module: msticpy.context.azure
            Class: MicrosoftSentinel
            InitArgs:
            Connect: True
            ConnectArgs:
                workspace: CyberSecuritySoc
                auth_methods: ['cli', 'device_code']

Example usage:

.. code-block:: python

    import msticpy as mp
    mp.init_notebook()
    mp.mp_user_session.load_user_config()

"""

from __future__ import annotations

import importlib
import logging
import sys
from pathlib import Path
from typing import Any

import yaml
from IPython import get_ipython

from .._version import VERSION
from ..data.core.data_providers import QueryProvider

__version__ = VERSION
__author__ = "Ian Hellen"

logger = logging.getLogger(__name__)


def load_user_session(
    session_file: str | Path | None = None,
    namespace: dict[str, Any] | None = None,
    verbose: bool = False,
):
    """
    Load user session configuration.

    Parameters
    ----------
    session_file : str, Path, optional
        Path to the user configuration file, by default None.
        If no file is provided, the function will look in the current directory
        for an `mp_user_session.yaml` file.
    namespace : dict[str, Any], optional
        Namespace to load the configuration into, by default None.
        If no namespace is provided, the configuration will be loaded into the
        global namespace of the caller (usually the notebook).
    verbose : bool, optional
        If True, set the logger to INFO level for the duration of the function,
        by default False.

    """
    current_level = logger.level
    if verbose:
        if current_level > logging.INFO:
            logger.setLevel(logging.INFO)
    try:
        if session_file is None:
            session_file = Path.cwd() / "mp_user_session.yaml"
        session_file = Path(session_file)
        if not session_file.exists():
            raise FileNotFoundError(f"Session file {session_file} not found.")

        if namespace is None:
            ipython = get_ipython()
            if ipython is None:
                # pylint: disable=protected-access
                namespace = sys._getframe(1).f_locals
            else:
                namespace = ipython.user_ns

        user_config = yaml.safe_load(session_file.read_text(encoding="utf-8"))
        _load_query_providers(user_config, namespace)
        _load_mp_components(user_config, namespace)
    finally:
        logger.setLevel(current_level)


def _load_query_providers(user_config, namespace):
    """
    Load and Initialize the query provider settings.

    Parameters
    ----------
    user_config : dict[str, Any]
        The user configuration settings.
    namespace : dict[str, Any]
        The namespace to load the component instances into.

    """
    logger.info(
        "Loading %d query providers", len(user_config.get("QueryProviders", {}))
    )
    for qry_prov_name, qry_prov_settings in user_config.get(
        "QueryProviders", {}
    ).items():
        qry_prov = _initialize_component(
            qry_prov_name, qry_prov_settings, QueryProvider
        )
        if qry_prov:
            namespace[qry_prov_name] = qry_prov


def _load_mp_components(user_config, namespace):
    """
    Load and Initialize non-query components.

    Parameters
    ----------
    user_config : dict[str, Any]
        The user configuration settings.
    namespace : dict[str, Any]
        The namespace to load the component instances into.

    """
    logger.info("Loading %d components", len(user_config.get("QueryProviders", {})))
    for comp_name, comp_settings in user_config.get("Components", {}).items():
        module = importlib.import_module(comp_settings.get("Module"), package="msticpy")
        comp_class = getattr(module, comp_settings.get("Class"))
        comp_instance = _initialize_component(comp_name, comp_settings, comp_class)
        if comp_instance:
            namespace[comp_name] = comp_instance


def _initialize_component(name, settings, cls) -> Any:
    """
    Initialize a component or query provider.

    Parameters
    ----------
    name : str
        The name of the component or query provider.
    settings : dict
        The settings for the component or query provider.
    cls : type
        The class of the component or query provider.

    Returns
    -------
    object
        The initialized component or query provider instance, or None if initialization failed.

    """
    try:
        init_args = settings.get("InitArgs", {})
        if init_args is None:
            init_args = {}
        if data_env := settings.get("DataEnvironment"):
            init_args["data_environment"] = data_env
        instance = cls(**init_args)
        logger.info("Initialized %s", name)
    # pylint: disable=broad-except
    except Exception:
        logger.exception("Failed to initialize %s", name, exc_info=True)
        return None

    if settings.get("Connect"):
        print(f"Connecting to {name}")
        try:
            connect_args = settings.get("ConnectArgs", {})
            if connect_args is None:
                connect_args = {}
            instance.connect(**connect_args)
            logger.info("Connected to %s", name)
        # pylint: disable=broad-except
        except Exception:
            logger.exception("Failed to connect to %s", name, exc_info=True)
            return None
    else:
        logger.info("Connect not requested for %s", name)

    return instance



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Facade module for settings functions."""

from .._version import VERSION

# pylint: disable=unused-import
from .pkg_config import (  # noqa: F401
    current_config_path,
    get_config,
    get_http_timeout,
    get_settings,
    refresh_config,
    set_config,
    validate_config,
)
from .provider_settings import (  # noqa: F401
    clear_keyring,
    get_protected_setting,
    get_provider_settings,
    refresh_keyring,
    reload_settings,
)
from .proxy_settings import get_http_proxies  # noqa: F401

__version__ = VERSION
__author__ = "Ian Hellen"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for SQL to KQL Conversion.

This is an experiment conversion utility built to support a limited subset
of ANSI SQL.

It relies on mo_sql_parsing https://github.com/klahnakoski/mo-sql-parsing
(a maintained fork from the deprecated https://github.com/mozilla/moz-sql-parser)
to parse the SQL syntax tree. Some hacky additions have been done to
allow table renaming and support for non ANSI SQL operators such as
RLIKE.

For a more complete translation help with SQL to KQL see
https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sqlcheatsheet

Known limitations
-----------------

- Does not support aggregate functions in SELECT with no GROUP BY clause
- Does not support IN, EXISTS, HAVING operators
- Only partial support for AS naming (should work in SELECT expressions)

"""

import random
import re
from collections import namedtuple
from typing import Any, Dict, List, Optional, Tuple, Union

from ..common.exceptions import MsticpyImportExtraError

try:
    from mo_sql_parsing import parse
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without mo-sql-parsing installed",
        title="Error importing mo_sql_parsing for sql_to_kql",
        extra="sql",
    ) from imp_err


from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

_DEBUG = False

FuncFormat = namedtuple("FuncFormat", "default, cust_arg_fmt, cust_func_format")

SQL_KQL_FUNC_MAP = {
    "add": FuncFormat(None, None, "{p0} + {p1}"),
    "avg": FuncFormat("avg", None, None),
    "base64": FuncFormat("base64_encode_tostring", None, None),
    "concat": FuncFormat("strcat", None, None),
    "count": FuncFormat("count", None, None),
    "if": FuncFormat("iif", None, None),
    "iif": FuncFormat("iif", None, None),
    "ifnull": FuncFormat(None, None, "iif(isnull({p0}), {p1}, {p0}))"),
    "in": FuncFormat("in", None, None),
    "instr": FuncFormat("indexof", None, None),
    "int": FuncFormat("toint", None, None),
    "isnotnull": FuncFormat("isnotnull", None, None),
    "isnull": FuncFormat("isnull", None, None),
    "left": FuncFormat("NA", None, None),
    "length": FuncFormat("strlen", None, None),
    "like": FuncFormat("NA", None, None),
    "locate": FuncFormat("indexof", None, None),
    "lower": FuncFormat("tolower", None, None),
    "lcase": FuncFormat("tolower", None, None),
    "ltrim": FuncFormat("trim_start", None, None),
    "max": FuncFormat("max", None, None),
    "mean": FuncFormat("mean", None, None),
    "min": FuncFormat("min", None, None),
    "position": FuncFormat("indexof", None, None),
    "regexp_extract": FuncFormat("extract", "{p1}, {p0}", None),  # swap params 0, 1
    "replace": FuncFormat("replace", None, None),
    "reverse": FuncFormat("reverse", None, None),
    "rtrim": FuncFormat("trim_end", None, None),
    "split": FuncFormat("split", None, None),
    "string": FuncFormat("tostring", None, None),
    "str": FuncFormat("tostring", None, None),
    "substr": FuncFormat("substring", None, None),
    "substring": FuncFormat("substring", None, None),
    "sum": FuncFormat("sum", None, None),
    "todatetime": FuncFormat("to_date", None, None),
    "translate": FuncFormat("translate", None, None),
    "trim": FuncFormat("trim", None, None),
    "unbase64": FuncFormat("base64_decode_tostring", None, None),
    "upper": FuncFormat("toupper", None, None),
}


AND = "and"
AS = "as"
ASC = "asc"
BETWEEN = "between"
CASE = "case"
COLLATE_NOCASE = "collate nocase"
CROSS_JOIN = "cross join"
DESC = "desc"
DISTINCT = "distinct"
ELSE = "else"
END = "end"
FROM = "from"
FULL_JOIN = "full join"
FULL_OUTER_JOIN = "full outer join"
GROUP_BY = "groupby"
HAVING = "having"
IN = "in"
INNER_JOIN = "inner join"
IS = "is"
IS_NOT = "is not"
JOIN = "join"
LEFT_JOIN = "left join"
LEFT_OUTER_JOIN = "left outer join"
LIKE = "like"
LIMIT = "limit"
NOT = "not"
NOT_BETWEEN = "not_between"
NOT_IN = "nin"
NOT_LIKE = "not_like"
NOT_RLIKE = "not_rlike"
OFFSET = "offset"
ON = "on"
OR = "or"
ORDER_BY = "orderby"
RIGHT_JOIN = "right join"
RIGHT_OUTER_JOIN = "right outer join"
RLIKE = "rlike"
SELECT = "select"
SELECT_DISTINCT = "select_distinct"
THEN = "then"
UNION = "union"
UNION_ALL = "union all"
USING = "using"
WHEN = "when"
WHERE = "where"
WITH = "with"

JOIN_KEYWORDS = {
    FULL_JOIN: "outer",
    FULL_OUTER_JOIN: "outer",
    INNER_JOIN: "inner",
    JOIN: "inner",
    LEFT_JOIN: "left",
    LEFT_OUTER_JOIN: "left",
    RIGHT_JOIN: "right",
    RIGHT_OUTER_JOIN: "right",
    CROSS_JOIN: "CROSS_JOIN_TODO",
}
JOIN_KEYWORDS = {kw.replace("_", " "): kql for kw, kql in JOIN_KEYWORDS.items()}

BINARY_OPS = {
    "concat": "concat",
    "mul": "*",
    "div": "/",
    "mod": "%",
    "add": "+",
    "sub": "-",
    "binary_and": "binary_and",
    "binary_or": "binary_or",
    "binary_xor": "binary_xor",
    "lt": "<",
    "lte": "<=",
    "gt": ">",
    "gte": ">=",
    "not_between": "not_between",
    "or": "or",
    "and": "and",
    "eq": "==",
    "neq": "!=",
    "nin": "!in",
    "rlike": "matches regex",
    "not_rlike": "not matches regex",
    "is": "==",
    "is_not": "!=",
}

# noqa: MC0001


def sql_to_kql(sql: str, target_tables: Dict[str, str] = None) -> str:
    """Parse SQL and return KQL equivalent."""
    # ensure literals are surrounded by single quotes
    sql = _single_quote_strings(sql)

    # replace table names
    if target_tables:
        for table in target_tables:
            sql = sql.replace(table, target_tables[table])
    # replace keywords
    # sql = _remap_kewords(sql)
    parsed_sql = parse(sql)
    query_lines = _parse_query(parsed_sql)
    return "\n".join(line for line in query_lines if line.strip())


def _parse_query(parsed_sql: Dict[str, Any]) -> List[str]:  # noqa: MC0001
    """Translate query or subquery."""
    query_lines: List[str] = []
    if isinstance(parsed_sql, str):
        return [parsed_sql]
    if FROM in parsed_sql:
        _process_from(parsed_sql[FROM], query_lines)
    if WHERE in parsed_sql:
        query_lines.append(f"| where {_parse_expression(parsed_sql[WHERE])}")

    if GROUP_BY in parsed_sql:
        _process_group_by(parsed_sql, query_lines)
        # Get rid of the SELECT statement since we've processed it in
        # the groupby
        parsed_sql.pop(SELECT, parsed_sql.pop(SELECT_DISTINCT, None))

    distinct_select: List[Dict[str, Any]] = []
    if SELECT_DISTINCT in parsed_sql:
        distinct_select.extend(parsed_sql[SELECT_DISTINCT])
        _process_select(
            parsed_sql[SELECT_DISTINCT], parsed_sql[SELECT_DISTINCT], query_lines
        )
    if SELECT in parsed_sql:
        _process_select(parsed_sql[SELECT], parsed_sql[SELECT], query_lines)
    if ORDER_BY in parsed_sql:
        query_lines.append(f"| order by {_create_order_by(parsed_sql[ORDER_BY])}")
    if distinct_select:
        query_lines.append(
            f"| distinct {', '.join(_create_distinct_list(distinct_select))}"
        )
    if LIMIT in parsed_sql:
        query_lines.append(f"| limit {parsed_sql[LIMIT]}")
    if UNION in parsed_sql:
        union_subquery = {UNION_ALL: parsed_sql[UNION]}
        query_lines.extend(_parse_query(union_subquery))
        query_lines.append("| distinct *")
    if UNION_ALL in parsed_sql:
        union_l_expr = "\n".join(_parse_query(parsed_sql[UNION_ALL][0]))
        query_lines.append(union_l_expr)
        union_r_expr = "\n  ".join(_parse_query(parsed_sql[UNION_ALL][1]))
        query_lines.append(f"| union ({union_r_expr}\n)")

    return query_lines


def _process_from(
    from_expr: Union[List[Dict[str, Any]], Dict[str, Any], str], query_lines: List[str]
):
    """Process FROM clause."""
    if isinstance(from_expr, dict) and UNION in from_expr:
        query_lines.extend(_parse_query(from_expr))
    elif isinstance(from_expr, dict):
        query_lines.extend(_parse_query(from_expr))
    elif isinstance(from_expr, str):
        query_lines.append((from_expr))
        return
    elif isinstance(from_expr, list):
        for from_item in from_expr:
            if isinstance(from_item, str):
                query_lines.append((from_item))
            elif isinstance(from_item, dict) and "value" in from_item:
                query_lines.extend(_parse_query(from_item.get("value")))  # type: ignore

    join_expr = from_expr if isinstance(from_expr, list) else [from_expr]
    join_list = _get_join_list(join_expr)
    for join_item in join_list:
        join_line = _parse_join(join_item)
        if join_line:
            query_lines.append(join_line)


def _process_select(
    parsed_sql: Dict[str, Any],
    expr_list: Union[List[Dict[Any, Any]], Dict[Any, Any]],
    query_lines: List[str],
):
    """Process SELECT clause."""
    # Expressions
    if parsed_sql == "*":
        return
    _db_print(expr_list, type(expr_list))
    select_list = expr_list if isinstance(expr_list, list) else [expr_list]
    select_list = _get_expr_list(select_list)
    project_items = []
    extend_items = []
    for item in select_list:
        value = _parse_expression(item["value"])
        name = item.get("name")
        if value != item["value"]:
            # if this isn't a simple rename - add to extend
            name = name or _gen_expr_name(item["value"])
            extend_items.append(f"{name} = {value}")
            project_items.append(name)
        else:
            if name:
                project_items.append(f"{name} = {value}")
            else:
                project_items.append(value)
    if extend_items:
        query_lines.append(f"| extend {', '.join(extend_items)}")
    if project_items:
        query_lines.append(f"| project {', '.join(project_items)}")


def _gen_expr_name(value):
    """Generate random expression name."""
    pref = "expr"
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        first_val = next(iter(value.values()))
        if isinstance(first_val, str):
            return first_val
        # prob a function so base the name on that
        pref = next(iter(value.keys()))
    # otherwise just generate a rand value as suffix
    suffix = str(random.randint(1000, 9999))  # nosec
    return f"{pref}_{suffix}"


def _get_expr_list(expr_list):
    if (
        isinstance(expr_list, list)
        and len(expr_list) == 1
        and isinstance(expr_list[0], dict)
        and isinstance(expr_list[0].get("value"), list)
    ):
        return expr_list[0]["value"]
    return expr_list


def _get_expr_value(expr_val):
    if isinstance(expr_val, dict) and "value" in expr_val:
        return expr_val["value"]
    if isinstance(expr_val, list):
        return _get_expr_list(expr_val)
    return expr_val


def _process_group_by(parsed_sql: Dict[str, Any], query_lines: List[str]):
    """Process GROUP BY clause."""
    group_by_expr = parsed_sql[GROUP_BY]
    group_by_expr = (
        group_by_expr if isinstance(group_by_expr, list) else [group_by_expr]
    )
    by_clause = ", ".join(val["value"] for val in group_by_expr if val.get("value"))

    expr_list = parsed_sql.get(SELECT, parsed_sql.get(SELECT_DISTINCT, []))
    group_by_expr_list = []
    expr_list = _get_expr_value(expr_list)
    for expr in expr_list:
        name_expr = ""
        if "name" in expr:
            name_expr = f"{expr.get('name')} = "
        if isinstance(expr.get("value"), str):
            group_by_expr_list.append(f"{name_expr}any({expr['value']})")
        else:
            group_by_expr_list.append(
                f"{name_expr}{_parse_expression(expr.get('value'))}"
            )
    query_lines.append(f"| summarize {', '.join(group_by_expr_list)} by {by_clause}")


# pylint: disable=too-many-return-statements, too-many-branches
def _parse_expression(expression):  # noqa: MC0001
    """Return parsed expression."""
    if _is_literal(expression)[0]:
        return _quote_literal(expression)
    if not isinstance(expression, dict):
        return expression
    if AND in expression:
        return "\n  and ".join(
            [f"({_parse_expression(expr)})" for expr in expression[AND]]
        )
    if OR in expression:
        return "\n  or ".join(
            [f"({_parse_expression(expr)})" for expr in expression[OR]]
        )
    if NOT in expression:
        return f" not ({_parse_expression(expression[NOT])})"
    if BETWEEN in expression:
        args = expression[BETWEEN]
        betw_expr = f"{_parse_expression(args[1])} .. {_parse_expression(args[2])}"
        return f"{args[0]} between ({betw_expr})"
    if NOT_BETWEEN in expression:
        args = expression[NOT_BETWEEN]
        betw_expr = f"{_parse_expression(args[1])} .. {_parse_expression(args[2])}"
        return f"{args[0]} !between ({betw_expr})"
    if IN in expression or NOT_IN in expression:
        sql_op = IN if IN in expression else NOT_IN
        kql_op = IN if IN in expression else "!in"
        args = expression[sql_op]

        right = _quote_literal(args[1])
        if isinstance(right, list):
            _db_print(args[1])
            arg_list = ", ".join([str(_parse_expression(l_item)) for l_item in right])
            return f"{args[0]} {kql_op} ({arg_list})"
        sub_query = "\n".join(_parse_query(right))
        return f"{args[0]} {kql_op} ({sub_query})"

    # Handle other operators
    oper = next(iter(expression.keys())) if expression else None
    if oper in BINARY_OPS:
        right = _parse_expression(expression[oper][1])
        left = _parse_expression(expression[oper][0])
        return f"{left} {BINARY_OPS[oper]} {right}"
    if LIKE in expression:
        return _process_like(expression)

    # For everything else, assume it's a function
    if expression:
        func, operand = next(iter(expression.items()))
        #         set_trace()
        if isinstance(operand, list):
            return _map_func(func, *operand)
        return _map_func(func, operand)
    return "EXPRESSION {expression} not resolved."


# pylint: enable=too-many-return-statements, too-many-branches


def _map_func(func: str, *args) -> str:
    """Return KQL function for SQL function."""
    func = func.lower().strip()
    args_dict = {f"p{idx}": arg for idx, arg in enumerate(args)}
    def_arg_fmt = ", ".join(f"{{{arg}}}" for arg in args_dict)
    if func not in SQL_KQL_FUNC_MAP:
        func_fmt = f"{func}({def_arg_fmt}) // WARNING unmapped function\n"
        return func_fmt.format(**args_dict)
    func_map = SQL_KQL_FUNC_MAP[func]

    if (
        func == "count"
        and isinstance(args[0], dict)
        and next(iter(args[0])) == "distinct"
    ):
        func_arg = _get_expr_value(args[0]["distinct"])
        return f"dcount({func_arg})"

    if not func_map.cust_arg_fmt and not func_map.cust_func_format:
        func_fmt = f"{func_map.default}({def_arg_fmt})"
        return func_fmt.format(**args_dict)
    if func_map.cust_arg_fmt:
        func_fmt = f"{func_map.default}({func_map.cust_arg_fmt})"
        return func_fmt.format(**args_dict)
    if func_map.cust_func_format:
        func_fmt = f"{func_map.cust_func_format}"
        return func_fmt.format(**args_dict)
    raise ValueError(f"Could not map function or args {func}{args_dict}")


def _quote_literal(expr: Union[str, List[str], Any]) -> Any:
    """Quote string if it is a literal."""
    literal, expr = _is_literal(expr)
    if not literal:
        return expr
    if isinstance(expr, str):
        return _quote(expr)
    if isinstance(expr, list):
        return [_quote(memb) for memb in expr]
    return expr


def _is_literal(expr: Union[Dict[str, Any], Any]) -> Tuple[bool, Any]:
    """Check if literal string."""
    if isinstance(expr, dict) and "literal" in expr:
        return True, expr["literal"]
    return False, expr


def _quote(expr: str) -> str:
    """Quote a string, if not already quoted."""
    if expr.startswith("'") and expr.endswith("'"):
        return expr
    return f"'{expr}'"


def _single_quote_strings(sql: str) -> str:
    """Replace unquoted double-quotes with single-quotes."""
    return re.sub(r"(?<![\\])\"", "'", sql)


def _format_order_item(item: Dict[str, Any]) -> str:
    """Return ORDER BY item with sort direction."""
    if "sort" in item:
        return f"{item['value']} {item['sort'].lower()}"
    return f"{item['value']}"


def _get_join_list(parsed_sql: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Return list of JOIN sub-expressions."""
    if not isinstance(parsed_sql, list):
        return []
    join_list = []
    for from_source in parsed_sql:
        if not isinstance(from_source, dict):
            continue
        join = JOIN_KEYWORDS & from_source.keys()
        if join:
            join_list.append(from_source)
    return join_list


def _rewrite_table_refs(join_expr: Union[Any, str, List], table_expr: str) -> str:
    """Rewrite dotted prefixes."""
    p_expr = _parse_expression(join_expr)
    prefixes = set(re.findall(r"(\w+)\.", p_expr))
    if not prefixes:
        return p_expr
    if f"{table_expr}" in prefixes:
        p_expr = p_expr.replace(f"{table_expr}.", "$right.")
        prefixes.remove(f"{table_expr}")
    for prefix in prefixes:
        p_expr = p_expr.replace(
            f"{prefix}.",
            "$right." if prefix.casefold() == table_expr.casefold() else "$left.",
        )
    return p_expr


def _parse_join(join_expr) -> Optional[str]:
    """Return translated JOIN expression."""
    join_type_set = JOIN_KEYWORDS & join_expr.keys()
    if not join_type_set:
        return None
    join_type = join_type_set.pop()
    table_expr = join_expr[join_type]
    kql_join_type = JOIN_KEYWORDS[join_type]
    if "value" in table_expr and "select" in table_expr["value"]:
        table_expr = table_expr["value"]

    p_table_expr = "\n  ".join(_parse_query(table_expr))
    if "name" in join_expr[join_type]:
        table_name = join_expr[join_type]["name"]
    else:
        table_name = p_table_expr.split(" ", maxsplit=1)[0].strip()
    on_expr = _parse_expression(join_expr["on"])
    on_expr = _rewrite_table_refs(on_expr, table_name)
    _db_print(table_expr, kql_join_type, p_table_expr)

    return f"| join kind={kql_join_type} ({p_table_expr}) on {on_expr}"


def _process_like(expression: Dict[str, Any]) -> str:
    """Process Like clause."""
    left = _parse_expression((expression[LIKE][0]))
    literal, right = _is_literal(expression[LIKE][1])
    if not (literal and isinstance(right, str)):
        raise ValueError(
            f"Right side operand {right} isn't usable in LIKE expression",
            f"{left} LIKE {right}",
        )
    if re.match("^[^%_]+[%_]$", right):
        oper = "startswith"
        right = right.replace("%", "").replace("_", "")
    elif re.match("^[%_][^%_]+$", right):
        oper = "endswith"
        right = right.replace("%", "").replace("_", "")
    elif re.match("^[%_][^%_]+[%_]$", right):
        oper = "contains"
        right = right.replace("%", "").replace("_", "")
    else:
        oper = "matches regex"
        right = right.replace("_", ".").replace("%", ".*")
    right = _quote(right)
    return f"{left} {oper} {right}"


def _create_distinct_list(distinct_select):
    distinct_list = []
    for distinct_item in distinct_select:
        if "name" in distinct_item:
            distinct_list.append(distinct_item["name"])
        else:
            val = _parse_expression(_get_expr_value(distinct_item))
            if val != distinct_item.get("value"):
                # If value was a complex expression we can't use
                # it directly so just revert to distinct *
                distinct_list = ["*"]
                break
            distinct_list.append(val)
    return distinct_list


def _create_order_by(order_by):
    if isinstance(order_by, list):
        return ", ".join(_format_order_item(item) for item in order_by)
    return _format_order_item(order_by)


def _db_print(*args, **kwargs):
    if _DEBUG:
        print(*args, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Analytics Features."""
from __future__ import annotations

import datetime as dt
import logging
from typing import TYPE_CHECKING, Any

import httpx
from azure.common.exceptions import CloudError
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import (
    SentinelUtilsMixin,
    _azs_api_result_to_df,
    extract_sentinel_response,
    get_http_timeout,
)

if TYPE_CHECKING:
    import pandas as pd

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)
_INDICATOR_ITEMS: dict[str, str] = {
    "valid_to": "validUntil",
    "description": "description",
    "threat_types": "threatTypes",
    "name": "displayName",
    "confidence": "confidence",
    "revoked": "revoked",
    "source": "source",
}

_IOC_TYPE_MAPPING: dict[str, str] = {
    "dns": "domain-name",
    "url": "url",
    "ipv4": "ipv4-addr",
    "ipv6": "ipv6-addr",
    "md5_hash": "MD5",
    "sha1_hash": "SHA-1",
    "sha256_hash": "SHA-256",
}

MAX_CONFIDENCE: int = 100


class SentinelTIMixin(SentinelUtilsMixin):
    """Mixin class for Sentinel Hunting feature integrations."""

    def get_all_indicators(
        self: Self,
        limit: int | None = None,
        orderby: str | None = None,
    ) -> pd.DataFrame:
        """
        Return all TI indicators in a Microsoft Sentinel workspace.

        Parameters
        ----------
        limit: int, optional
            If set returns top n results
        orderby : Optional[str], optional
            Order results by a specific column

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        appendix = "/indicators?"
        if limit:
            appendix += f"&$top={limit}"
        if orderby:
            appendix += f"&$orderby={orderby}"
        return self._list_items(
            item_type="ti",
            api_version="2021-10-01",
            appendix=appendix,
        )

    def get_ti_metrics(self: Self) -> pd.DataFrame:
        """
        Return metrics about TI indicators in a Microsoft Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        return self._list_items(
            item_type="ti",
            api_version="2021-10-01",
            appendix="/metrics",
        )

    def create_indicator(  # pylint:disable=too-many-arguments, too-many-locals #noqa:PLR0913
        self: Self,
        indicator: str,
        ioc_type: str,
        name: str = "TI Indicator",
        confidence: int = 0,
        *,
        silent: bool = False,
        description: str | None = None,
        labels: list | None = None,
        kill_chain_phases: list | None = None,
        threat_types: list | None = None,
        external_references: list | None = None,
        valid_from: dt.datetime | None = None,
        valid_to: dt.datetime | None = None,
    ) -> str:
        """
        Create a new indicator within the Microsoft Sentinel workspace.

        Parameters
        ----------
        indicator : str
            The indicator to create - i.e. IP address, domain name etc.
        ioc_type : str
            The type of indicator to create - can be:
            "dns", "url", "ipv4", "ipv6", "md5_hash", "sha1_hash", "sha256_hash"
        name : str, optional
            A common name to give to the indicator default is 'TI Indicator'
        confidence : int, optional
            A score between 0-100 of the confidence in the indicator, defualt is 0
        silent : bool, optional
            If True no output is displayed, defaults to False
        description : str, optional
            An description of the indicator
        labels : list, optional
            A list of string object labels to associate with the indicator
        kill_chain_phases : list, optional
            A list of string objects relating to the kill chain phases
            an indicator is assocaited with
        threat_types : list, optional
            A list of threat types associated with the indicator (list of string objects)
        external_references : list, optional
            A list of URLs that provide an external reference for the indicator
        valid_from : datetime, optional
            A datetime from which the indicator is valid from, defaults to now
        valid_to : datetime, optional
            A datetime to which the indicator is valid until

        Returns
        -------
        The ID of the created indicator

        Raises
        ------
        MsticpyUserError
            If invalid ioc_type or confidence value provided
        CloudError
            If API call fails

        """
        self.check_connected()
        ti_url: str = self.sent_urls["ti"] + "/createIndicator"
        params: dict[str, str] = {"api-version": "2021-10-01"}
        if ioc_type not in _IOC_TYPE_MAPPING:
            err_msg: str = """ioc_type must be one of -
                 'dns', 'url', 'ipv4', 'ipv6', 'md5_hash', 'sha1_hash', 'sha256_hash'"""
            raise MsticpyUserError(err_msg)
        normalized_ioc_type: str = _IOC_TYPE_MAPPING[ioc_type]
        pattern_type: str = normalized_ioc_type
        value: str = "value"
        if normalized_ioc_type in ["SHA-256", "SHA-1", "MD5"]:
            pattern_type = "file"
            value = f"hashes.'{normalized_ioc_type}'"

        if confidence > MAX_CONFIDENCE or confidence < 0:
            err_msg = "confidence must be between 0 and 100"
            raise MsticpyUserError(err_msg)
        data_items: dict[str, Any] = {
            "displayName": name,
            "confidence": confidence,
            "pattern": f"[{pattern_type}:{value} = '{indicator}']",
            "patternType": pattern_type,
            "revoked": "false",
            "source": "MSTICPy",
        }
        data_items.update(
            _build_additional_indicator_items(
                valid_from=valid_from,
                valid_to=valid_to,
                external_references=external_references,
                kill_chain_phases=kill_chain_phases,
                labels=labels,
                name=name,
                threat_types=threat_types,
                confidence=confidence,
                description=description,
            ),
        )
        data: dict[str, Any] = extract_sentinel_response(data_items, props=True)
        data["kind"] = "indicator"
        if not self._token:
            err_msg = "Token not found, can't create indicator."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.post(
            ti_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        if not silent:
            logger.info("Indicator created.")

        return response.json().get("name")

    def bulk_create_indicators(
        self: Self,
        data: pd.DataFrame,
        indicator_column: str = "Observable",
        indicator_type_column: str = "IoCType",
        *,
        confidence_column: str | None = None,
    ) -> None:
        """
        Bulk create indicators from a DataFrame.

        Parameters
        ----------
        data : pd.DataFrame
            A dataframe containing indicators and indicator types
        indicator_column : str, optional
            The column containing indicator values to create, by default "Observable"
        indicator_type_column : str, optional
            The column containing indicator type values, by default "IoCType"
        confidence_column : str, optional
            The column containing indicator confidence values, by default 0 value used.

        """
        for row in data.iterrows():
            confidence: int = row[1][confidence_column] if confidence_column else 0
            try:
                self.create_indicator(
                    indicator=row[1][indicator_column],
                    ioc_type=row[1][indicator_type_column],
                    confidence=confidence,
                    silent=True,
                )
            except CloudError:
                logger.exception(
                    "Error creating indicator %s",
                    row[1][indicator_column],
                )
        logger.info("%s indicators created.", len(data.index))

    def get_indicator(self: Self, indicator_id: str) -> dict:
        """
        Get a specific indicator by its ID.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to get

        Returns
        -------
        dict
            Indicator details

        Raises
        ------
        CloudError
            If API call fails.

        """
        self.check_connected()
        ti_url: str = self.sent_urls["ti"] + f"/indicators/{indicator_id}"
        params: dict[str, str] = {"api-version": "2021-10-01"}
        if not self._token:
            err_msg = "Token not found, can't get indicator."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.get(
            ti_url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        return response.json()

    def update_indicator(  # pylint:disable=too-many-arguments,too-many-locals #noqa:PLR0913
        self: Self,
        indicator_id: str,
        *,
        name: str | None = None,
        confidence: int = 0,
        description: str | None = None,
        labels: list[str] | None = None,
        kill_chain_phases: list | None = None,
        threat_types: list | None = None,
        external_references: list | None = None,
        valid_from: dt.datetime | None = None,
        valid_to: dt.datetime | None = None,
    ) -> None:
        """
        Update an existing indicator within the Microsoft Sentinel workspace.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to update
        name : str, optional
            A common name to give to the indicator default is 'TI Indicator'
        confidence : int, optional
            A score between 0-100 of the confidence in the indicator
        description : str, optional
            An description of the indicator
        labels : list, optional
            A list of string object labels to associate with
            the indicator
        kill_chain_phases : list, optional
            A list of string objects relating to the kill chain phases
            an indicator is assocaited with
        threat_types : list, optional
            A list of threat types associated with the indicator (list of string objects)
        external_references : list, optional
            A list of URLs that provide an external reference for the indicator
        valid_from : datetime, optional
            A datetime from which the indicator is valid from, defaults to now
        valid_to : datetime, optional
            A datetime to which the indicator is valid until

        Raises
        ------
        CloudError
            If API call fails

        """
        self.check_connected()
        ti_url: str = self.sent_urls["ti"] + f"/indicators/{indicator_id}"
        indicator_details: dict[str, Any] = self.get_indicator(indicator_id)
        data_items: dict[str, Any] = _build_additional_indicator_items(
            valid_from=valid_from,
            valid_to=valid_to,
            external_references=external_references,
            kill_chain_phases=kill_chain_phases,
            labels=labels,
            name=name,
            threat_types=threat_types,
            confidence=confidence,
            description=description,
        )
        data_items.pop("validFrom")
        full_data_items: dict[str, Any] = _add_missing_items(
            data_items,
            indicator_details,
        )
        data: dict[str, Any] = extract_sentinel_response(full_data_items, props=True)
        data["etag"] = indicator_details["etag"]
        data["kind"] = "indicator"
        params: dict[str, str] = {"api-version": "2021-10-01"}
        if not self._token:
            err_msg = "Token not found, can't update indicator."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            ti_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        logger.info("Indicator updated.")

    def add_tag(self: Self, indicator_id: str, tag: str) -> None:
        """
        Add a tag to an existing indicator.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to add a tag to.
        tag : str
            The tag to add.

        """
        self.check_connected()
        indicator_details: dict[str, Any] = self.get_indicator(indicator_id)
        tags: list[str] = [tag]
        if "threatIntelligenceTags" in indicator_details["properties"]:
            tags += indicator_details["properties"]["threatIntelligenceTags"]
        self.update_indicator(indicator_id=indicator_id, labels=tags)

    def delete_indicator(self: Self, indicator_id: str) -> None:
        """
        Delete a specific TI indicator.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to delete

        Raises
        ------
        CloudError
            If API call fails

        """
        self.check_connected()
        ti_url: str = self.sent_urls["ti"] + f"/indicators/{indicator_id}"
        params: dict[str, str] = {"api-version": "2021-10-01"}
        if not self._token:
            err_msg = "Token not found, can't delete indicator."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.delete(
            ti_url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 204):
            raise CloudError(response=response)
        logger.info("Indicator deleted.")

    def query_indicators(  # pylint:disable=too-many-arguments, too-many-locals #noqa:PLR0913
        self: Self,
        *,
        include_disabled: bool = False,
        keywords: str | None = None,
        min_confidence: int = 0,
        max_confidence: int = 100,
        max_valid_until: str | None = None,
        min_valid_until: str | None = None,
        page_size: int | None = None,
        pattern_types: list[str] | None = None,
        sort_by: list[str] | None = None,
        sources: list[str] | None = None,
        threat_types: list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Query for indicators in a Sentinel workspace.

        Parameters
        ----------
        include_disabled : bool, optional
            Parameter to include/exclude disabled indicators.
        keywords : str, optional
            Keyword for searching threat intelligence indicators
            Use this to search for specific indicator values.
        max_confidence : int, optional
            Maximum confidence.
        max_valid_until : str, optional
            End time for ValidUntil filter.
        min_confidence : int, optional
            Minimum confidence.
        min_valid_until : str, optional
            Start time for ValidUntil filter.
        page_size : int, optional
            Maximum number of results to return in one page.
        pattern_types : list, optional
            A list of IoC types to include.
        sort_by : List, optional
            Columns to sort by and sorting order as:
            [{"itemKey": COLUMN_NAME, "sortOrder": ascending/descending}]
        sources: list, optional
            A list of indicator sources to include
        threat_types: list, optional
            A list of Threat types to include

        Returns
        -------
        pd.DataFrame
            A set of matching indicators

        Raises
        ------
        CloudError
            If API call fails

        """
        self.check_connected()
        ti_url: str = self.sent_urls["ti"] + "/queryIndicators"
        data_items: dict[str, Any] = {
            "includeDisabled": include_disabled,
            "maxConfidence": max_confidence,
            "minConfidence": min_confidence,
        }
        if keywords:
            data_items["keywords"] = keywords
        if max_valid_until:
            data_items["maxValidUntil"] = max_valid_until
        if min_valid_until:
            data_items["minValidUntil"] = min_valid_until
        if page_size:
            data_items["pageSize"] = page_size
        if pattern_types:
            data_items["patternTypes"] = pattern_types
        if sort_by:
            data_items["sortBy"] = sort_by
        if sources:
            data_items["sources"] = sources
        if threat_types:
            data_items["threatTypes"] = threat_types
        params: dict[str, str] = {"api-version": "2021-10-01"}
        if not self._token:
            err_msg = "Token not found, can't query indicators."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.post(
            ti_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data_items),
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        return _azs_api_result_to_df(response)


def _build_additional_indicator_items(  # pylint:disable=too-many-arguments #noqa: PLR0913
    *,
    valid_from: dt.datetime | None = None,
    valid_to: dt.datetime | None = None,
    external_references: list[str] | None = None,
    kill_chain_phases: list[str] | None = None,
    labels: list[str] | None = None,
    name: str | None = None,
    confidence: int = 0,
    description: str | None = None,
    threat_types: list | None = None,
    revoked: bool | None = None,
    source: str | None = None,
) -> dict:
    """Add in additional data items for indicators."""
    data_items: dict[str, Any] = {
        "validFrom": (
            valid_from.isoformat()
            if valid_from
            else dt.datetime.now(tz=dt.timezone.utc).isoformat()
        ),
        "confidence": confidence,
    }
    if name:
        data_items["displayName"] = name
    if description:
        data_items["description"] = description
    if threat_types:
        data_items["threatTypes"] = threat_types
    if revoked:
        data_items["revoked"] = revoked
    if source:
        data_items["source"] = source
    if valid_to:
        data_items["validUntil"] = valid_to.isoformat()
    else:
        data_items["validUntil"] = dt.datetime.now(tz=dt.timezone.utc).isoformat()
    if external_references:
        ext_refs: list[dict[str, Any]] = [
            {"sourceName": "MSTICPy", "url": ref} for ref in external_references
        ]
        data_items["externalReferences"] = ext_refs
    if kill_chain_phases:
        kill_chain: list[dict[str, Any]] = [
            {
                "killChainName": "Lockheed Martin - Intrusion Kill Chain",
                "phaseName": phase,
            }
            for phase in kill_chain_phases
        ]
        data_items["killChainPhases"] = kill_chain
    if labels:
        data_items["labels"] = labels
        data_items["threatIntelligenceTags"] = labels
    return data_items


_REQUIRED_ITEMS: list[str] = ["pattern", "patternType", "source"]


def _add_missing_items(
    data_items: dict[str, Any],
    indicator: dict[str, Any],
) -> dict[str, Any]:
    """Add missing required items to requests based on existing values."""
    for req_item in _REQUIRED_ITEMS:
        if req_item not in data_items:
            data_items[req_item] = indicator["properties"][req_item]
    return data_items



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Github Sentinel Query repo import class and helpers."""

import logging
import os
import re
import warnings
import zipfile
from datetime import datetime
from pathlib import Path
from typing import Optional

import attr
import httpx
import yaml
from attr import attrs
from tqdm.notebook import tqdm

from ..._version import VERSION

__version__ = VERSION
__author__ = "Jannie Li"

# pylint: disable=too-many-instance-attributes
# pylint: disable=consider-using-with
# pylint: disable=too-many-locals
# pylint: disable=unspecified-encoding


QUERY_METADATA_SECTION = [
    "name",
    "severity",
    "tags",
    "required_data_connectors",
    "query_frequency",
    "query_period",
    "trigger_operator",
    "trigger_threshold",
    "tactics",
    "relevant_techniques",
    "entity_mappings",
    "custom_details",
    "alert_details_override",
    "version",
    "kind",
    "folder_name",
    "source_file_name",
    "query_type",
]


QUERY_DEFAULT_PARAMETER_SECTION = {
    "add_query_items": {
        "description": "Additional query clauses",
        "type": "str",
        "default": "",
    },
    "start": {"description": "Query start time", "type": "datetime"},
    "end": {"description": "Query end time", "type": "datetime"},
}


@attrs(auto_attribs=True)
class SentinelQuery:
    """Attrs class that represents a Sentinel Query yaml file."""

    query_id: str = attr.Factory(str)
    name: str = attr.Factory(str)
    description: str = attr.Factory(str)
    severity: str = attr.Factory(str)
    query_frequency: str = attr.Factory(str)
    query_period: str = attr.Factory(str)
    trigger_operator: str = attr.Factory(str)
    trigger_threshold: str = attr.Factory(str)
    version: str = attr.Factory(str)
    kind: str = attr.Factory(str)
    folder_name: str = attr.Factory(str)
    source_file_name: str = attr.Factory(str)
    query_type: str = attr.Factory(str)
    tactics: list = attr.Factory(list)
    relevant_techniques: list = attr.Factory(list)
    query: str = attr.Factory(str)
    entity_mappings: dict = attr.Factory(dict)
    custom_details: dict = attr.Factory(dict)
    alert_details_override: dict = attr.Factory(dict)
    tags: list = attr.Factory(list)
    required_data_connectors: dict = attr.Factory(dict)


def get_sentinel_queries_from_github(
    git_url: Optional[
        str
    ] = "https://github.com/Azure/Azure-Sentinel/archive/master.zip",
    outputdir: Optional[str] = None,
) -> bool:
    r"""
    Download Microsoft Sentinel Github archive and extract detection and hunting queries.

    Parameters
    ----------
    git_url : str, optional
        URL of the GIT Repository to be downloaded, by default
        "https://github.com/Azure/Azure-Sentinel/archive/master.zip"
    outputdir : str, optional
        Provide absolute path to the output folder to save downloaded archive
        (e.g. '/usr/home' or 'C:\downloads').
        If no path provided, it will download to .msticpy dir under Azure-Sentinel directory.

    """
    if outputdir is None:
        outputdir = str(
            Path.joinpath(Path("~").expanduser(), ".msticpy", "Azure-Sentinel")
        )

    try:
        with httpx.stream("GET", git_url, follow_redirects=True) as response:  # type: ignore
            progress_bar = tqdm(
                desc="Downloading from Microsoft Sentinel Github",
                initial=0,
                unit="iB",
                unit_scale=True,
            )
            repo_zip = Path.joinpath(Path(outputdir), "Azure-Sentinel.zip")  # type: ignore
            with open(repo_zip, "wb") as file:
                for data in response.iter_bytes(chunk_size=10000):
                    progress_bar.update(len(data))
                    file.write(data)
            progress_bar.close()

            archive = zipfile.ZipFile(repo_zip, mode="r")

        # Only extract Detections and Hunting Queries Folder
        for file in archive.namelist():  # type: ignore
            if file.startswith(  # type: ignore
                (
                    "Azure-Sentinel-master/Detections/",
                    "Azure-Sentinel-master/Hunting Queries/",
                )
            ):
                archive.extract(file, path=outputdir)  # type: ignore
        print("Downloaded and Extracted Files successfully")
        return True

    except httpx.HTTPError as http_err:
        warnings.warn(f"HTTP error occurred trying to download from Github: {http_err}")
        return False


def read_yaml_files(parent_dir: str, child_dir: str) -> dict:
    """
    Create dictionary mapping query file paths with the yaml file text each contains.

    Parameters
    ----------
    parent_dir : str
        Directory storing the Hunting and Detections directories
    child_dir : str
        Either "Hunting Queries" or "Detections" or otherwise named query category

    Returns
    -------
    dict
       Dictionary mapping query file paths to corresponding yaml file text in the
       parent_dir/child_dir specified. Only identifies .yaml files.

    """
    # enumerate the files and read the yaml
    yaml_queries = list(Path(parent_dir, child_dir).rglob("*.yaml"))
    yaml_queries_list = [str(Path(q)) for q in yaml_queries]
    parsed_query_dict = {}

    for query in yaml_queries_list:
        with open(query, encoding="utf8", errors="ignore") as opened_query_file:
            parsed_query_dict[query] = opened_query_file.read()

    return parsed_query_dict


def import_sentinel_queries(yaml_files: dict, query_type: str) -> list:
    """
    Create list of SentinelQuery attr objects.

    Parameters
    ----------
    yaml_files : dict
        Dictionary mapping query file addresses to yaml file text created by read_yaml_files
    query_type : str
        Either "Hunting Queries" or "Detections" or otherwise named query category

    Returns
    -------
    list
        Returns a list of SentinelQuery attr objects from a dict of yaml files and query type given

    """
    return [
        _import_sentinel_query(yaml_path, yaml_text, query_type)
        for yaml_path, yaml_text in yaml_files.items()
    ]


def _import_sentinel_query(
    yaml_path: str, yaml_text: str, query_type: str
) -> SentinelQuery:
    """
    Create a SentinelQuery attr object for a given yaml query.

    Parameters
    ----------
    yaml_path : str
        File path to a YAML Sentinel query
    yaml_text : str
        YAML text (Sentinel query) associated with the yaml_path
    query_type : str
        Either "Hunting Queries" or "Detections" or otherwise named query category

    Returns
    -------
    SentinelQuery
        Returns an attrs object called SentinelQuery with all the YAML query information

    """
    logger = logging.getLogger(__name__)
    try:
        parsed_yaml_dict = yaml.load(yaml_text, Loader=yaml.SafeLoader)
    except yaml.YAMLError as error:
        logger.warning("Failed to parse yaml for %s", yaml_path)
        logger.warning(error)
        return SentinelQuery()

    new_query = SentinelQuery(
        name=parsed_yaml_dict.get("name"),
        query_id=parsed_yaml_dict.get("id", ""),
        description=parsed_yaml_dict.get("description", ""),
        severity=parsed_yaml_dict.get("severity", ""),
        tags=parsed_yaml_dict.get("tags_entry", []),
        required_data_connectors=parsed_yaml_dict.get("requiredDataConnectors", {}),
        query_frequency=parsed_yaml_dict.get("queryFrequency", ""),
        query_period=parsed_yaml_dict.get("queryPeriod", ""),
        trigger_operator=parsed_yaml_dict.get("triggerOperator", ""),
        trigger_threshold=parsed_yaml_dict.get("triggerThreshold", ""),
        tactics=parsed_yaml_dict.get("tactics", ""),
        relevant_techniques=parsed_yaml_dict.get("relevantTechniques", []),
        query=parsed_yaml_dict.get("query", ""),
        version=parsed_yaml_dict.get("version", ""),
        kind=parsed_yaml_dict.get("kind", ""),
        folder_name=yaml_path.replace("\\", "/").split("/")[-2],
        source_file_name=yaml_path,
        query_type=query_type,
    )
    return new_query


def _format_query_name(qname: str) -> str:
    """
    Format the inputted query name for use as a file name.

    Parameters
    ----------
    qname : str
        Name of a Sentinel Query as read from the corresponding YAML file

    Returns
    -------
    str
        Returns the inputted string with non-alphanumeric characters removed and
        spaces replaced with an underscore

    """
    del_chars_pattern = r"""[%*\*'\"\\.()[\]{}-]"""
    repl_str = re.sub(del_chars_pattern, "", qname)
    repl_str = repl_str.replace(" ", "_").replace("__", "_")
    return repl_str


def _organize_query_list_by_folder(query_list: list) -> dict:
    """
    Create a dictionary mapping each folder name with related SentinelQuery objects.

    Parameters
    ----------
    query_list : list
        List of SentinelQuery objects returned by import_sentinel_queries()

    Returns
    -------
    dict
        Returns a dictionary mapping each folder name with the SentinelQuery objects associated
        with it

    """
    queries_by_folder = {}
    for query in query_list:
        if query.folder_name == "":
            warnings.warn(f"query {query} has no folder_name")
        if query.folder_name not in queries_by_folder:
            queries_by_folder[query.folder_name] = [query]
        else:
            queries_by_folder[query.folder_name].append(query)

    return queries_by_folder


def _create_queryfile_metadata(folder_name: str) -> dict:  # type: ignore
    """
    Generate metadata section of the YAML file for the given folder_name.

    Parameters
    ----------
    folder_name : str
        Either "Hunting Queries" or "Detections" or otherwise named query category

    Returns
    -------
    dict
        Returns a generated metadata section for the YAML files in the given folder_name

    """
    dict_to_write = {"metadata": {}, "defaults": {}, "sources": {}}  # type: ignore
    dict_to_write["metadata"]["version"] = 1  # write update version functionality
    dict_to_write["metadata"]["description"] = "Sentinel Alert Queries - " + folder_name
    dict_to_write["metadata"]["data_environments"] = [
        "MSSentinel"
    ]  # write how to get this and what type this should be
    dict_to_write["metadata"]["data_families"] = folder_name
    dict_to_write["metadata"]["last_updated"] = str(datetime.now())
    return dict_to_write


def _create_yaml_source_sec(cur_query: dict) -> dict:
    """
    Create the metadata section of the YAML for the current query.

    Parameters
    ----------
    cur_query : SentinelQuery
        The current SentinelQuery attrs object that a metadata section will be generated for.

    Returns
    -------
    dict
        Returns generated metadata section of the YAML for the given individual query.

    """
    source_dict = {}
    source_dict["description"] = cur_query["description"]
    source_dict["metadata"] = {}
    source_dict["metadata"]["sentinel"] = {"query_id": cur_query["query_id"]}
    for section in QUERY_METADATA_SECTION:
        source_dict["metadata"][section] = cur_query[section]
    source_dict["metadata"]["args"] = {}
    source_dict["metadata"]["args"]["query"] = cur_query["query"]
    source_dict["metadata"]["parameters"] = {}
    return source_dict


def write_to_yaml(query_list: list, query_type: str, output_folder: str) -> bool:
    """
    Write out generated YAML files of the given query_list into the given output_folder.

    Parameters
    ----------
    query_list : list
        List of SentinelQuery attr objects generated by import_sentinel_queries()
    query_type : str
        Either "Hunting Queries" or "Detections" or otherwise named query category
    output_folder : str
        The name of the folder you want the written YAML files to be stored in

    Returns
    -------
    bool
        True if succeeded; False if an error occurred

    """
    logger = logging.getLogger(__name__)
    query_dict_by_folder = _organize_query_list_by_folder(query_list)
    all_folders = query_dict_by_folder.keys()

    for source_folder in all_folders:
        logger.info("now writing files from %s", source_folder)
        dict_to_write = _create_queryfile_metadata(source_folder)

        if query_type == "Detections":
            dict_to_write["defaults"]["parameters"] = QUERY_DEFAULT_PARAMETER_SECTION

        for cur_query in query_dict_by_folder[source_folder]:
            # skipping instances where there is no name but should probably have a better solution
            try:
                formatted_qname = _format_query_name(cur_query.name)
                dict_to_write["sources"][formatted_qname] = _create_yaml_source_sec(
                    attr.asdict(cur_query)
                )

            except TypeError as err:
                logger.warning(
                    """Query name is most likely None at %s for
                    current query %r""",
                    source_folder,
                    cur_query,
                )
                print(err)

        try:
            query_text = yaml.safe_dump(
                dict_to_write, encoding="utf-8", sort_keys=False
            )
        except yaml.YAMLError as error:
            print(error)
            return False

        try:
            def_path = Path.joinpath(Path(os.getcwd()))
            path_main = Path(def_path, output_folder)
            path_type = Path(output_folder, query_type)
            if not path_main.is_dir():
                path = Path(def_path, output_folder)
                os.mkdir(path)
            if not path_type.is_dir():
                path = Path(output_folder, query_type)
                os.mkdir(path)
            Path(output_folder, query_type, source_folder).write_text(
                query_text.decode("utf-8")
            )
        except OSError as error:
            print(error)
            return False
    logger.info("done writing files")
    return True


def download_and_write_sentinel_queries(
    query_type: str, yaml_output_folder: str, github_outputdir: Optional[str] = None
):
    """
    Download queries from GitHub and write out YAML files for the given query type.

    Parameters
    ----------
    query_type : str
        Either "Hunting Queries" or "Detections" or otherwise named query category
    yaml_output_folder : str
        Path to the folder you want the new generated YAML files to be stored in
    github_outputdir : Optional[str]
        Path to the directory you want the Github download to be stored in

    """
    print("Downloading files from GitHub")
    get_sentinel_queries_from_github(outputdir=github_outputdir)
    print("Reading yaml_files")
    if github_outputdir is None:
        github_outputdir = str(
            Path.joinpath(Path("~").expanduser(), ".msticpy", "Azure-Sentinel")
        )
    base_dir = str(Path(github_outputdir, "/Azure-Sentinel-master"))
    yaml_files = read_yaml_files(parent_dir=base_dir, child_dir=query_type)
    print("Generating a list of queries")
    query_list = import_sentinel_queries(yaml_files, query_type=query_type)
    query_list = [
        query for query in query_list if query.query_id != ""
    ]  # may need better solution for failed query definitions
    print("Writing to YAML output folder")
    write_to_yaml(query_list, query_type, yaml_output_folder)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
from datetime import datetime, timedelta
from typing import Optional

import ipywidgets as widgets
from ipywidgets import Layout

from .._version import VERSION
from .core import (
    IPyDisplayMixin,
    default_before_after,
    default_max_buffer,
    parse_time_unit,
)

__version__ = VERSION
__author__ = "Ian Hellen"


class Lookback(IPyDisplayMixin):
    """Time lookback slider."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        default: Optional[int] = None,
        description: str = "Select time ({units}) to look back",
        origin_time: datetime = None,
        min_value: Optional[int] = None,
        max_value: Optional[int] = None,
        units: str = "hour",
        auto_display: bool = False,
        **kwargs,
    ):
        """
        Create an instance of the lookback slider widget.

        Parameters
        ----------
        default : int, optional
            The default 'lookback' time (the default is 4)
        description : str, optional
            The description to display
            (the default is 'Select time ({units}) to look back')
        origin_time : datetime, optional
            The origin time (the default is `datetime.utcnow()`)
        min_value : int, optional
            Minimum value (the default is 1)
        max_value : int, optional
            Maximum value (the default is 240)
        units : str, optional
            Time unit (the default is 'hour')
            Permissable values are 'day', 'hour', 'minute', 'second',
            'week'
            These can all be abbreviated down to initial characters
            ('d', 'm', etc.)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        # default to now
        self.origin_time = datetime.utcnow() if origin_time is None else origin_time
        description = kwargs.pop("label", description)

        self._time_unit = parse_time_unit(units)
        if "{units}" in description:
            description = description.format(units=self._time_unit.name)
        default = default_before_after(default, self._time_unit)
        min_value = min_value or 1
        max_value = default_max_buffer(max_value, default, self._time_unit)

        self._lookback_wgt = widgets.IntSlider(
            value=default,
            min=min_value,
            max=max_value,
            step=1,
            description=description,
            layout=Layout(width="60%", height="50px"),
            style={"description_width": "initial"},
        )

        self.end = self.origin_time
        self._time_unit = parse_time_unit(units)
        self.start = self.end - timedelta(
            seconds=(self._time_unit.value * self._lookback_wgt.value)
        )

        self._lookback_wgt.observe(self._time_range_change, names="value")

        if auto_display:
            self.display()

    @property
    def layout(self):
        """Return underlying widget."""
        return self._lookback_wgt

    @property
    def lookback(self):
        """Return current widget lookback value."""
        return self._lookback_wgt.value

    @property
    def value(self):
        """Return current widget lookback value."""
        return self._lookback_wgt.value

    def _time_range_change(self, change):
        del change
        self.start = self.origin_time - timedelta(
            0, self._lookback_wgt.value * self._time_unit.value
        )
